#! /usr/bin/env python
#
"""
Unit tests for the `gc3libs.backends.shellcmd` module.
"""
# Copyright (C) 2011-2015, 2018  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases()
from builtins import object
__docformat__ = 'reStructuredText'


from io import StringIO
import os
import shutil
import tempfile
import time

import pytest

import gc3libs
from gc3libs.backends.shellcmd import ShellcmdLrms
import gc3libs.config
import gc3libs.core
from gc3libs.quantity import Duration, Memory, s, kB
from gc3libs.testing.helpers import (
    SuccessfulApp,
    temporary_core,
    temporary_directory,
)


class TestBackendShellcmd(object):
    CONF = """
[resource/localhost_test]
type=shellcmd
transport=local
# time_cmd=/usr/bin/time
# Use unusual values so that we can easily spot if the `override` option works
max_cores=123
max_cores_per_job=123
max_memory_per_core=999
max_walltime=2
architecture=x64_64
auth=noauth
enabled=True
override=False
resourcedir=%s

[auth/noauth]
type=none
"""

    @pytest.fixture(autouse=True)
    def setUp(self):
        (fd, cfgfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        CONFIG = TestBackendShellcmd.CONF % (cfgfile + '.d')
        f.write(CONFIG)
        f.close()
        self.files_to_remove = [cfgfile, cfgfile + '.d']
        self.apps_to_kill = []

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(cfgfile)

        self.core = gc3libs.core.Core(self.cfg)
        self.backend = self.core.get_backend('localhost_test')
        # Update resource status
        self.backend.get_resource_status()

        yield

        for fname in self.files_to_remove:
            if os.path.isdir(fname):
                shutil.rmtree(fname)
            elif os.path.exists(fname):
                os.remove(fname)

        for app in self.apps_to_kill:
            try:
                self.core.kill(app)
            except:
                pass
            try:
                self.core.free(app)
            except:
                pass

    def cleanup_file(self, fname):
        self.files_to_remove.append(fname)

    def run_until_terminating(self, app, max_wait=10, polling_interval=0.1):
        """
        Wait until the given `app` task is done,
        but timeout and raise an error if it takes too much time.
        """
        waited = 0
        while (app.execution.state != gc3libs.Run.State.TERMINATING
                and waited < max_wait):
            time.sleep(polling_interval)
            waited += polling_interval
            self.core.update_job_state(app)


    def test_submission_ok(self):
        """Test a successful submission cycle and the backends' resource book-keeping"""
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')

        app = gc3libs.Application(
            arguments=['/usr/bin/env'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            stdout="stdout.txt",
            stderr="stderr.txt",
            requested_cores=1, )
        self.core.submit(app)
        self.apps_to_kill.append(app)

        self.cleanup_file(tmpdir)
        self.cleanup_file(app.execution.lrms_execdir)

        # there's no SUBMITTED state here: jobs go immediately into
        # RUNNING state
        assert app.execution.state == gc3libs.Run.State.SUBMITTED
        assert self.backend.free_slots == 122
        assert self.backend.user_queued == 0
        assert self.backend.user_run == 1

        self.run_until_terminating(app)
        try:
            assert app.execution.state == gc3libs.Run.State.TERMINATING
            assert self.backend.free_slots == 123
            assert self.backend.user_queued == 0
            assert self.backend.user_run == 0
        except:
            self.core.fetch_output(app)
            self.core.free(app)
            raise

        self.core.fetch_output(app)
        try:
            assert app.execution.state == gc3libs.Run.State.TERMINATED
            assert self.backend.free_slots == 123
            assert self.backend.user_queued == 0
            assert self.backend.user_run == 0
        except:
            self.core.free(app)
            raise

    def test_check_app_after_reloading_session(self):
        """Check that the job status is still available the end of the starter script"""

        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        self.cleanup_file(tmpdir)

        app = gc3libs.Application(
            arguments=['/usr/bin/env'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            stdout="stdout.txt",
            stderr="stderr.txt",
            requested_cores=1, )
        self.core.submit(app)
        self.apps_to_kill.append(app)

        self.cleanup_file(app.execution.lrms_execdir)

        # The wrapper process should die and write the final status
        # and the output to a file, so that `Core` will be able to
        # retrieve it.
        self.run_until_terminating(app)
        assert app.execution.state == gc3libs.Run.State.TERMINATING
        assert app.execution.returncode == 0

    def test_app_argument_with_spaces(self):
        """Check that arguments with spaces are not split"""
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        self.cleanup_file(tmpdir)

        app = gc3libs.Application(
            arguments=['/bin/ls', '-d', '/ /'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            stdout="stdout.txt",
            stderr="stderr.txt",
            requested_cores=1, )
        self.core.submit(app)
        self.apps_to_kill.append(app)
        self.cleanup_file(app.execution.lrms_execdir)

        self.run_until_terminating(app)
        assert app.execution.state == gc3libs.Run.State.TERMINATING
        assert app.execution.returncode != 0

    def test_resource_usage(self):
        """Check book-keeping of core and memory resources"""
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        self.cleanup_file(tmpdir)

        app = gc3libs.Application(
            arguments=['/bin/echo', 'Hello', 'World'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            requested_cores=2,
            requested_memory=10 * Memory.MB, )
        cores_before = self.backend.free_slots
        mem_before = self.backend.available_memory
        self.core.submit(app)
        self.apps_to_kill.append(app)

        cores_after = self.backend.free_slots
        mem_after = self.backend.available_memory
        assert cores_before == cores_after + 2
        assert mem_before == mem_after + app.requested_memory
        self.run_until_terminating(app)
        assert self.backend.free_slots == cores_before
        assert self.backend.available_memory == mem_before

    def test_env_vars_definition(self):
        """Check that `Application.environment` settings are correctly propagated"""
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        self.cleanup_file(tmpdir)

        app = gc3libs.Application(
            arguments=['/bin/echo', '$MSG'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            stdout='stdout.txt',
            environment={'MSG':'OK'})
        self.core.submit(app)
        self.apps_to_kill.append(app)

        self.run_until_terminating(app)

        self.core.fetch_output(app)
        stdout_file = os.path.join(app.output_dir, app.stdout)
        assert os.path.exists(stdout_file)
        assert os.path.isfile(stdout_file)
        stdout_contents = open(stdout_file, 'r').read()
        assert stdout_contents == 'OK\n'

    def test_stdout_in_directory(self):
        """Check that `Application.stdout` can include a full path"""
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        self.cleanup_file(tmpdir)

        app = gc3libs.Application(
            arguments=['/bin/echo', 'OK'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            stdout='logs/stdout.txt')
        self.core.submit(app)
        self.apps_to_kill.append(app)

        self.run_until_terminating(app)

        self.core.fetch_output(app)
        stdout_dir = os.path.join(app.output_dir, 'logs')
        stdout_file = os.path.join(stdout_dir, 'stdout.txt')
        assert os.path.exists(stdout_dir)
        assert os.path.isdir(stdout_dir)
        assert os.path.exists(stdout_file)
        assert os.path.isfile(stdout_file)
        stdout_contents = open(stdout_file, 'r').read()
        assert stdout_contents == 'OK\n'

    def test_not_enough_cores_usage(self):
        """Check that a `NoResources` exception is raised if more cores are requested than available"""
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        self.cleanup_file(tmpdir)
        bigapp = gc3libs.Application(
            arguments=['/bin/echo', 'Hello', 'World'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            requested_cores=self.backend.free_slots + 1,
            requested_memory=10 * Memory.MiB, )
        with pytest.raises(gc3libs.exceptions.NoResources):
            self.core.submit(bigapp)

    def test_not_enough_memory_usage(self):
        """Check that a `NoResources` exception is raised if more memory is requested than available"""
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        self.cleanup_file(tmpdir)
        bigapp = gc3libs.Application(
            arguments=['/bin/echo', 'Hello', 'World'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            requested_cores=1,
            requested_memory=self.backend.total_memory + Memory.B, )
        with pytest.raises(gc3libs.exceptions.NoResources):
            self.core.submit(bigapp)


class TestBackendShellcmdCFG(object):
    CONF = """
[resource/localhost_test]
type=shellcmd
transport=local
time_cmd=/usr/bin/time
max_cores=1000
max_cores_per_job=4
max_memory_per_core=2
max_walltime=2
architecture=x64_64
auth=noauth
enabled=True
override=%s
resourcedir=%s

[auth/noauth]
type=none
"""

    @pytest.fixture(autouse=True)
    def setUp(self):
        self.files_to_remove = []

        yield
        for fname in self.files_to_remove:
            if os.path.isdir(fname):
                shutil.rmtree(fname)
            elif os.path.exists(fname):
                os.remove(fname)

    def cleanup_file(self, fname):
        self.files_to_remove.append(fname)


    def test_override_cfg_flag(self):
        (fd, cfgfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendShellcmdCFG.CONF % ("True", cfgfile + '.d'))
        f.close()
        self.files_to_remove = [cfgfile, cfgfile + '.d']

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(cfgfile)

        self.core = gc3libs.core.Core(self.cfg)
        self.backend = self.core.get_backend('localhost_test')
        # Update resource status
        self.backend.get_resource_status()

        assert self.backend.max_cores < 1000

    def test_do_not_override_cfg_flag(self):
        (fd, cfgfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendShellcmdCFG.CONF % ("False", cfgfile + '.d'))
        f.close()
        self.files_to_remove = [cfgfile, cfgfile + '.d']

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(cfgfile)

        self.core = gc3libs.core.Core(self.cfg)
        self.backend = self.core.get_backend('localhost_test')
        # Update resource status

        assert self.backend.max_cores == 1000

    def test_resource_sharing_w_multiple_backends(self):
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        (fd, cfgfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendShellcmdCFG.CONF % ("False", cfgfile + '.d'))
        f.close()
        self.files_to_remove = [cfgfile, cfgfile + '.d', tmpdir]

        cfg1 = gc3libs.config.Configuration()
        cfg1.merge_file(cfgfile)

        cfg2 = gc3libs.config.Configuration()
        cfg2.merge_file(cfgfile)

        core1 = gc3libs.core.Core(cfg1)
        core2 = gc3libs.core.Core(cfg2)

        backend1 = core1.get_backend('localhost_test')
        backend2 = core2.get_backend('localhost_test')

        app = gc3libs.Application(
            arguments=['/bin/echo', 'Hello', 'World'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            requested_cores=1,
            requested_memory=10 * Memory.MiB, )

        try:
            core1.submit(app)
            assert (backend1.free_slots ==
                         backend1.max_cores - app.requested_cores)

            assert backend2.free_slots == backend2.max_cores
            backend2.get_resource_status()
            assert (backend2.free_slots ==
                         backend2.max_cores - app.requested_cores)
        finally:
            core1.kill(app)
            core1.free(app)


    def test_pid_list_is_string(self):
        tmpdir = tempfile.mkdtemp(prefix=__name__, suffix='.d')
        (fd, cfgfile) = tempfile.mkstemp()
        f = os.fdopen(fd, 'w+')
        f.write(TestBackendShellcmdCFG.CONF % ("False", cfgfile + '.d'))
        f.close()
        self.files_to_remove = [cfgfile, cfgfile + '.d']

        self.cfg = gc3libs.config.Configuration()
        self.cfg.merge_file(cfgfile)

        self.core = gc3libs.core.Core(self.cfg)
        self.backend = self.core.get_backend('localhost_test')
        # Update resource status

        app = gc3libs.Application(
        arguments=['/bin/echo', 'Hello', 'World'],
            inputs=[],
            outputs=[],
            output_dir=tmpdir,
            requested_cores=1,
            requested_memory=10 * Memory.MiB, )

        try:
            self.core.submit(app)
            for pid in list(self.backend._job_infos.keys()):
                assert isinstance(pid,str)
        finally:
            self.core.kill(app)
            self.core.free(app)

def test_shellcmd_backend_create_spooldir():
    with temporary_directory() as tmpdir:
        with temporary_core(spooldir=tmpdir) as core:
            assert 'test' in core.resources
            backend = core.get_backend('test')
            app = SuccessfulApp()
            core.submit(app)
            assert os.path.isdir(backend.spooldir)


def test_shellcmd_backend_parse_wrapper_output_successful():
    resource_usage = StringIO("""
WallTime=2.10s
KernelTime=0.61s
UserTime=0.67s
CPUUsage=61%
MaxResidentMemory=61604kB
AverageResidentMemory=0kB
AverageTotalMemory=0kB
AverageUnsharedMemory=0kB
AverageUnsharedStack=0kB
AverageSharedMemory=0kB
PageSize=4096B
MajorPageFaults=0
MinorPageFaults=101799
Swaps=0
ForcedSwitches=82
WaitSwitches=487
Inputs=0
Outputs=400
SocketReceived=0
SocketSent=0
Signals=0
ReturnCode=0
    """)
    with temporary_core(type='shellcmd') as core:
        assert 'test' in core.resources
        backend = core.get_backend('test')
        termstatus, acctinfo, valid = backend._parse_wrapper_output(resource_usage)
        assert valid
        assert acctinfo.duration == 2.10*s
        assert acctinfo.shellcmd_kernel_time == 0.61*s
        assert acctinfo.shellcmd_user_time == 0.67*s
        assert acctinfo.max_used_memory == 61604*kB
        assert acctinfo.shellcmd_average_total_memory == 0*kB
        assert acctinfo.shellcmd_filesystem_inputs == 0
        assert acctinfo.shellcmd_filesystem_outputs == 400
        assert acctinfo.shellcmd_swapped == 0
        assert termstatus == (0, 0)


def test_shellcmd_backend_parse_wrapper_output_signalled():
    resource_usage = StringIO("""Command terminated by signal 15
WallTime=6.44s
KernelTime=0.00s
UserTime=0.00s
CPUUsage=0%
MaxResidentMemory=2004kB
AverageResidentMemory=0kB
AverageTotalMemory=0kB
AverageUnsharedMemory=0kB
AverageUnsharedStack=0kB
AverageSharedMemory=0kB
PageSize=4096B
MajorPageFaults=0
MinorPageFaults=73
Swaps=0
ForcedSwitches=0
WaitSwitches=2
Inputs=0
Outputs=0
SocketReceived=0
SocketSent=0
Signals=0
ReturnCode=0
    """)
    with temporary_core(type='shellcmd') as core:
        assert 'test' in core.resources
        backend = core.get_backend('test')
        termstatus, acctinfo, valid = backend._parse_wrapper_output(resource_usage)
        assert valid
        assert acctinfo.duration == 6.44*s
        assert acctinfo.shellcmd_kernel_time == 0.*s
        assert acctinfo.shellcmd_user_time == 0.*s
        assert acctinfo.max_used_memory == 2004*kB
        assert acctinfo.shellcmd_average_total_memory == 0*kB
        assert acctinfo.shellcmd_filesystem_inputs == 0
        assert acctinfo.shellcmd_filesystem_outputs == 0
        assert acctinfo.shellcmd_swapped == 0
        assert termstatus == (15, -1)


if __name__ == "__main__":
    pytest.main(["-v", __file__])



#! /usr/bin/env python
#
#   gscr.py -- Front-end script for running ParRecoveryFun Matlab
#   function with a given combination of reference models.
#
#   Copyright (C) 2014, 2015  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gscr.py --help`` for program usage
instructions.

Input parameters consists of:
@param_folder
@data_folder
@sound_index
...

Options:
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2014-11-13:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gscr
    gscr.GscrScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
import random

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask


EXP_INDEX = [0,1]

## custom application class
class GscrApplication(Application):
    """
    Custom class to wrap the execution of the Matlab script.
    """
    application_name = 'scr_analysis'

    def __init__(self, param_file, data_file, **extra_args):

        self.output_dir = extra_args['output_dir']
        self.result_dir = extra_args['result_dir']

        inputs = dict()
        outputs = dict()
        executables = []

        # Check if binary to be executed is provided as part of input arguments
        if 'run_binary' in extra_args:
            inputs[os.path.abspath(extra_args["run_binary"])] = "estimate_DCM.m"

        arguments = "matlab -nodesktop -nosplash -nodisplay -r \"estimate_DCM " \
                    "%s %s results;quit;\"" % (os.path.basename(param_file),
                                          os.path.basename(data_file))

        inputs[param_file] = os.path.basename(param_file)
        inputs[data_file] = os.path.basename(data_file)

        # Set output
        outputs['results/'] = 'results/'

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gscr.log',
            join=True,
            executables = executables,
            **extra_args)


class GscrScript(SessionBasedScript):
    """
    For each param file (with '.mat' extension) found in the 'param folder',
    GscrScript extracts the corresponding index (from filename) and searches for
    the associated file in 'data folder'. For each pair ('param_file','data_file'),
    GscrScript generates execution Tasks.

    The ``gscr`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gscr`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GscrApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GscrApplication,
            )

    def setup_args(self):

        self.add_param('params', type=str,
                       help="Path to the param files. "
                       " Files with extension .m will be considered.")

        self.add_param('data', type=str,
                       help="Path to the data files. "
                       " Data files will be associated to param files "
                       "using their index number. Es. param file: "
                       " .DCM_s24_can_159e-4Hz_bi_depth2_newVBA.mat will "
                       "be associated to a data file: "
                       "tscr_soundexp_scbd24.mat (index is 24).")

    def setup_options(self):
        self.add_param("-b", "--binary", metavar="[STRING]",
                       dest="run_binary", default=None,
                       help="Location of the Matlab compiled binary "
                       "version of the ParRecoveryFun. Default: None.")

        self.add_param("-S", "--store_results", type=str, metavar="[STRING]",
                       dest="store_results", default=None,
                       help="Location where all results will be aggregated. "
                       "Default: (session folder).")

    def parse_args(self):
        """
        Check validity of input parameters and selected benchmark.
        """

        if not os.path.isdir(self.params.params):
            raise OSError("No such file or directory: %s ",
                          os.path.abspath(self.params.params))

        if not os.path.isdir(self.params.data):
            raise OSError("No such file or directory: %s ",
                          os.path.abspath(self.params.data))

        if self.params.run_binary:
            if not os.path.isfile(self.params.run_binary):
                raise gc3libs.exceptions.InvalidUsage("Estimate function binary "
                                                      " file %s not found"
                                                      % self.params.run_binary)

    def new_tasks(self, extra):
        """
        For each of the network data and for each of the selected benchmarks,
        create a GscrApplication.

        First loop the input files, then loop the selected benchmarks
        """
        tasks = []

        for (param,data,dcm_index) in self.pair_param_data(self.params.params,self.params.data):

            jobname = "SCR-%d" % (dcm_index)

            extra_args = extra.copy()
            extra_args['dcm_index'] = dcm_index

            if self.params.run_binary:
                extra_args['run_binary'] = self.params.run_binary

            extra_args['jobname'] = jobname

            if self.params.store_results:
                if not os.path.isdir(self.params.store_results):
                    os.makedirs(self.params.store_results)
                extra_args['result_dir'] = self.params.store_results
            else:
                extra_args['result_dir'] = self.params.output
                extra_args['result_dir'] = extra_args['result_dir'].replace('NAME', self.params.session)

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', 'SCR-%d' % (dcm_index))

            gc3libs.log.info("Creating GscrApplication with: %s %s %d",
                             param,
                             data,
                             dcm_index)
            tasks.append(GscrApplication(
                param,
                data,
                **extra_args))

        return tasks

    def pair_param_data(self, param, data):
        """
        Walk through 'param'
        for each .m extract index
        find corresponding data file in 'data'
        return triple: [param_file, data_file, index]
        Agreed filename structures and string pattern:
        param: DCM_s[index]_can_159e-4Hz_bi_depth2_newVBA.mat
        data: tscr_HRA_1_[index].mat
        pattern: match the first number occourrance in param file with the
        last number occurrance in data file.
        """
        # Generate index of data files
        # Data file example: tscr_HRA_1_12.mat
        data_files = dict()

        for data_file in os.listdir(data):
            if os.path.isfile(os.path.join(data,data_file)) and \
            re.findall(r'\d+',os.path.basename(data_file)) :
                index = re.findall(r'\d+',os.path.basename(data_file))[-1]
                data_files[index] = os.path.join(data,data_file)

        processes = []

        for param_file in os.listdir(param):
            if param_file.endswith('.mat') and \
               re.findall(r'\d+',os.path.basename(param_file)):
                # Valid Matlab parameter file found
                # DCM_s24_can_159e-4Hz_bi_depth2_newVBA.mat
                index = re.findall(r'\d+',os.path.basename(param_file))[0]

                # Search corresponding file in data folder
                if index in data_files:
                    # yield (param_file, data_files[index], index)
                    gc3libs.log.info("Found new parameter/data pair, index %s." %
                                     index)
                    processes.append((os.path.join(param,param_file),
                                      data_files[index],
                                      int(index)))
                else:
                    gc3libs.log.error("No data file associated to: %s index: %s" %
                                      (param_file, index))

        return processes



#! /usr/bin/env python
#
"""
Driver script for performing an global optimization over the parameter space.
"""
# Copyright (C) 2011, 2012  University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
__author__ = 'Riccardo Murri <riccardo.murri@uzh.ch>, Benjamin Jonen <benjamin.jonen@bf.uzh.ch>'
# summary of user-visible changes
__changelog__ = """

"""
__docformat__ = 'reStructuredText'

# std module imports
import numpy as np
import os
import re
import shutil
import sys
import time
import datetime

# Calling syntax one4all:
# -b ~/workspace/fpProj/model/base/ -x ~/workspace/fpProj/model/bin/forwardPremiumOut -e ~/workspace/fpProj/empirical/ -C 16 -NP 8 -xVars 'EA sigmaA' -xVarsDom '0.8 0.94 0.004 0.006' -sv info -t one4all --itermax 1 --countryList 'AU DE' -J 1000 -X i686 -r localhost --norm "fpSqr2" --initialPop "" -N


# Remove all files in curPath if -N option specified.
if __name__ == '__main__':
    import sys
    if '-N' in sys.argv:
        import os
        path2Pymods = os.path.join(os.path.dirname(__file__), '../')
        if not sys.path.count(path2Pymods):
            sys.path.append(path2Pymods)
        curPath = os.getcwd()
        filesAndFolder = os.listdir(curPath)
        if 'gParaSearch.csv' in filesAndFolder: # if another paraSearch was run in here before, clean up.
            shutil.rmtree(curPath)

if __name__ == "__main__":
    import gParaSearch


# gc3 library imports
import gc3libs
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript
from gc3libs.workflow import SequentialTaskCollection, ParallelTaskCollection
import gc3libs.utils

# import logger
import logbook

path2Pymods = os.path.join(os.path.dirname(__file__), '../')
if not sys.path.count(path2Pymods):
    sys.path.append(path2Pymods)

from forwardPremium import paraLoop_fp, GPremiumApplication
from supportGc3 import update_parameter_in_file, getParameter, getIndex
from supportGc3 import getLogger, wrapLogger, emptyFun


from pymods.classes.tableDict import tableDict

sys.path.append(os.path.join(os.path.dirname(__file__), '../generateResults/'))
from createOverviewTable import createOverviewTable
from analyzeOverviewTable import anaOne4eachPair, anaOne4eachCtry, anaOne4all, nlcOne4eachCtry, nlcOne4eachPair, nlcOne4all, plotPopOne4eachCtry
import combineOverviews
from difEvoKenPrice import *

# verbosity levels for logbook
##CRITICAL = 6
##ERROR = 5
##WARNING = 4
##NOTICE = 3
##INFO = 2
##DEBUG = 1
##NOTSET = 0

# Set up logger
#self.streamVerb = solverVerb
logger = wrapLogger(loggerName = 'gParaSearchLogger', streamVerb = 'INFO', logFile = os.path.join(os.getcwd(), 'gParaSearch.log'))


class gParaSearchDriver(SequentialTaskCollection):

    def __init__(self, pathToExecutable, pathToStageDir, architecture, baseDir, xVars, initialPop,
                 nPopulation, domain, solverVerb, problemType, pathEmpirical,
                 itermax, xConvCrit, yConvCrit,
                 makePlots, optStrategy, fWeight, fCritical, ctryList, analyzeResults, nlc, plot3dTable, combOverviews,
                 output_dir = '/tmp', **extra_args):

        '''
          pathToExecutable: Path to main executable. 
          pathToStageDir  : Path to directory in which the action takes place. Usually set to os.getcwd(). 
          architecture    : Set the architecture used. 
          baseDir         : Directory in which the input files are assembled. This directory is sent as input to the cluseter. 
          xVars           : Names of the x variables.
          initialPop      : The initial population if recovering from earlier failure. 
          nPopulation     : Population size. 
          domain          : The domain of the x variables. List of (lowerbound, upperbound) tuples. 
          solverVerb      : Verbosity of the solver. 
          problemType     : The problem type of the forward premium optimization. One4all, one4each etc. 
          pathEmpirical   : Path to the empirical results of the forward premium project. 
          itermax         : Maximum # of iterations of the solver. 
          xConvCrit       : Convergence criteria for x variables. 
          yConvCrit       : Convergence criteria for the y variables. 
          makePlots       : Make plots? 
          optStrategy     : The kind of differential evolution strategy to use. 
          fWeight         : 
          fCritical       : 
          ctryList        : The list of ctrys analyzed. 
          analyzeResults  : Function to analyze the output retrieved from the servers. 
          nlc             : Constraint function. 
          plot3dTable     : Function to generate 3d plots. 
          combOverviews   : Function to combine overviews. 
        '''

        logger.debug('entering gParaSearchDriver.__init__')

        # Set up initial variables and set the correct methods.
        self.pathToStageDir = pathToStageDir
        self.problemType = problemType
        self.pathToExecutable = pathToExecutable
        self.architecture = architecture
        self.baseDir = baseDir
        self.verbosity = solverVerb.upper()
        self.jobname = extra_args['jobname']
        self.ctryList = ctryList.split()
        self.xVars = xVars
        self.domain = domain
        self.n = len(self.xVars.split())
        self.extra_args = extra_args
        self.analyzeResults = analyzeResults
        self.plot3dTable    = plot3dTable
        self.combOverviews = combOverviews

        # Set solver options
        S_struct = {}
        S_struct['I_NP']         = int(nPopulation)
        S_struct['F_weight']     = float(fWeight)
        S_struct['F_CR']         = float(fCritical)
        S_struct['I_D']          = self.n
        S_struct['lowerBds']     = np.array([ element[0] for element in domain ], dtype = 'float64')
        S_struct['upperBds']     = np.array([ element[1] for element in domain ], dtype = 'float64')
        S_struct['I_itermax']    = int(itermax)
        S_struct['F_VTR']        = float(yConvCrit)
        S_struct['I_strategy']   = int(optStrategy)
        S_struct['I_plotting']   = int(makePlots)
        S_struct['xConvCrit']    = float(xConvCrit)
        S_struct['workingDir']   = pathToStageDir
        S_struct['verbosity']    = self.verbosity

        # Initialize solver
        self.deSolver = deKenPrice(S_struct)

        # Set constraint function in solver.
        self.deSolver.nlc = nlc

        # create initial task and register it
        if initialPop:
            self.deSolver.newPop = np.loadtxt(initialPop, delimiter = '  ')
        else:
            self.deSolver.newPop = self.deSolver.drawInitialSample()

        #self.deSolver.plotPopulation(self.deSolver.newPop)
        self.deSolver.I_iter += 1
        self.evaluator = gParaSearchParallel(self.deSolver.newPop, self.deSolver.I_iter, self.pathToExecutable, self.pathToStageDir,
                                             self.architecture, self.baseDir, self.xVars, self.verbosity, self.problemType, self.analyzeResults,
                                             self.ctryList, **self.extra_args)

        initial_task = self.evaluator

        SequentialTaskCollection.__init__(self, self.jobname, [initial_task])


    def __str__(self):
        return self.jobname

    def next(self, *args):
        logger.debug('entering gParaSearchDriver.next')

        self.changed = True
        newVals = self.evaluator.target(self.deSolver.newPop)
        self.deSolver.updatePopulation(self.deSolver.newPop, newVals)
        # Stats for initial population:
        self.deSolver.printStats()
        # make full overview table
        self.combOverviews(runDir = self.pathToStageDir, tablePath = self.pathToStageDir)

        # make plots
        if self.deSolver.I_plotting:
            self.deSolver.plotPopulation(self.deSolver.FM_pop)
            self.plot3dTable()

        if not self.deSolver.checkConvergence():
            self.deSolver.newPop = self.deSolver.evolvePopulation(self.deSolver.FM_pop)
            # Check constraints and resample points to maintain population size.
            self.deSolver.newPop = self.deSolver.enforceConstrReEvolve(self.deSolver.newPop)
            self.deSolver.I_iter += 1
            self.evaluator = gParaSearch.gParaSearchParallel(self.deSolver.newPop, self.deSolver.I_iter, self.pathToExecutable, self.pathToStageDir,
                                             self.architecture, self.baseDir, self.xVars, self.verbosity, self.problemType, self.analyzeResults,
                                             self.ctryList, **self.extra_args)
            self.add(self.evaluator)
        else:
            # post processing
            if self.deSolver.I_plotting:
                self.plot3dTable()

            open(os.path.join(self.pathToStageDir, 'jobDone'), 'w')
            # report success of sequential task
            self.execution.returncode = 0
            return Run.State.TERMINATED
        return Run.State.RUNNING


class gParaSearchParallel(ParallelTaskCollection, paraLoop_fp):

    def __str__(self):
        return self.jobname


    def __init__(self, inParaCombos, iteration, pathToExecutable, pathToStageDir, architecture, baseDir, xVars,
                 solverVerb, problemType, analyzeResults, ctryList, **extra_args):
        
        '''
          Generate a list of tasks and initialize a ParallelTaskCollection with them. 
          Uses paraLoop class to generate a list of (descriptions, substitutions for the input files). Descriptions are generated from
          variable names that are hard coded in this method right now. 
          Uses method generateTaskList to create a list of GPremiumApplication's which are invoked from a list of inputs (appropriately adjusted input files), 
          the output directory and some further settings for each run. 
          
          inParaCombos:      List of tuples defining the parameter combinations.
          iteration:         Current iteration number. 
          pathToExecutable:  Path to the executable (the external program to be called). 
          pathToStageDir:    Root path. Usually os.getcwd()
          architecture:      32 or 64 bit.
          baseDir:           Directory in which the input files are located. 
          xVars:             Names of the x variables. 
          solverVerb:        Logger verbosity. 
          problemType:       Forward premium specific flag to determine which case to look at. 
          analyzeResults:    Function to use to analyze the emerging output. 
          ctryList:          Forward premium specific list of ctrys to look at. 
        '''

        logger.debug('entering gParaSearchParalell.__init__')


        # Set up initial variables and set the correct methods.
        self.pathToStageDir = pathToStageDir
        self.problemType = problemType
        self.executable = pathToExecutable
        self.architecture = architecture
        self.baseDir = baseDir
        self.verbosity = solverVerb.upper()
        self.xVars = xVars
        self.n = len(self.xVars.split())
        self.analyzeResults = analyzeResults
        self.ctryList = ctryList
        self.iteration = iteration
        self.jobname = 'evalSolverGuess' + '-' + extra_args['jobname'] + '-' + str(self.iteration)
        self.extra_args = extra_args
        tasks = []

        # --- createJobs_x ---

        # Log activity
        cDate = datetime.date.today()
        cTime = datetime.datetime.time(datetime.datetime.now())
        dateString = '{0:04d}-{1:02d}-{2:02d}-{3:02d}-{4:02d}-{5:02d}'.format(cDate.year, cDate.month, cDate.day, cTime.hour, cTime.minute, cTime.second)
        logger.debug('Establishing parallel task on %s' % dateString)

        # Enter an iteration specific folder
        self.iterationFolder = os.path.join(self.pathToStageDir, 'Iteration-' + str(self.iteration))
        try:
            os.mkdir(self.iterationFolder)
        except OSError:
            print '%s already exists' % self.iterationFolder

        # save population to file
        np.savetxt(os.path.join(self.iterationFolder, 'curPopulation'), inParaCombos, delimiter = '  ')

        # Take the list of parameter combinations and translate them in a comma separated list of values for each variable to be fed into paraLoop file.
        # This can be done much more elegantly with ','.join() but it works...
        vals = []
        nVariables = range(len(inParaCombos[0]))
        for ixVar in nVariables:
            varValString = ''
            for ixParaCombo, paraCombo in enumerate(inParaCombos):
                ### Should make more precise string conversion.
                varValString += str(paraCombo[ixVar])
                if ixParaCombo < len(inParaCombos) - 1:
                    varValString += ', '
            vals.append( varValString )


        # Make problem specific adjustments to the paraLoop file.
        if self.problemType == 'one4all':
            print 'one4all'
            variables = ['Ctry', 'Ctry', 'EA', 'EB', 'sigmaA', 'sigmaB']
            groups    = [ 0, 0, 1, 1, 1, 1 ]
            groupRestrs = [ 'lowerTr', 'lowerTr', 'diagnol', 'diagnol', 'diagnol', 'diagnol' ]
            writeVals = [ ", ".join(self.ctryList), ", ".join(self.ctryList), vals[0], vals[0], vals[1], vals[1] ]
            self.variables = ['EA','sigmaA']
            self.paraCombos = inParaCombos
            paraFiles = [ 'input/markovA.in', 'input/markovB.in', 'input/parameters.in', 'input/parameters.in', 'input/parameters.in', 'input/parameters.in' ]
            paraFileRegex = [ 'space-separated', 'space-separated', 'bar-separated', 'bar-separated' , 'bar-separated' , 'bar-separated'  ]
            self.analyzeResults.tablePath = self.iterationFolder

        elif self.problemType == 'one4eachPair':
            print 'one4eachPair'
            # Check if EA or sigmaA are alone in the specified parameters. If so make diagnol adjustments
            writeVals = []
            if 'EA' in self.xVars and not 'EB' in self.xVars:
                variables = [ 'EA', 'EB' ]
                groups = [ '0', '0' ]
                groupRestrs = [ 'diagnol', 'diagnol' ]

                writeVals.append(vals[0])
                writeVals.append(vals[0])
                paraCombosEA = [  np.append(ele[0], ele[0]) for ele in inParaCombos ]
            if 'sigmaA' in self.xVars and not 'sigmaB' in self.xVars:
                variables.append( 'sigmaA')
                variables.append('sigmaB')
                groups.append( '0')
                groups.append('0')
                groupRestrs.append( 'diagnol')
                groupRestrs.append( 'diagnol' )
                writeVals.append(vals[1])
                writeVals.append(vals[1])
                paraCombosSigmaA = [  np.append(ele[1], ele[1]) for ele in inParaCombos ]

            # match ctry with val
            ctryVals = {}
            for ixCtry, ctry in enumerate(ctryList):
                ctryVals[ctry] = vals

            self.variables = variables

            # Prepare paraCombos matching to resulting table. Used in analyzeOverviewTable
            # !!! This should be dependent on problem type or on missing variables in xvars. !!!
            paraCombos = []
            for EA,sA in zip(paraCombosEA, paraCombosSigmaA):
                paraCombo = np.append(EA, sA)
                paraCombos.append(paraCombo)
            self.paraCombos = paraCombos
            paraFiles = [ 'input/parameters.in', 'input/parameters.in', 'input/parameters.in', 'input/parameters.in' ]
            paraFileRegex = [  'bar-separated', 'bar-separated' , 'bar-separated' , 'bar-separated'  ]

        elif self.problemType == 'one4eachCtry':
            print 'one4eachCtry'

            ctry1List  = []
            ctry2List  = []
            EAList     = []
            EBList     = []
            sigmaAList = []
            sigmaBList = []
            self.paraCombos = []

            ctryIndices = getIndex([len(ctryList), len(ctryList)], 'lowerTr')
            for ixCombo in range(len(inParaCombos)):
                ctry1ListCombo = []
                ctry2ListCombo = []
                EAListCombo    = []
                EBListCombo    = []
                sigmaAListCombo = []
                sigmaBListCombo = []
                for ctryIndex in ctryIndices:
                    ctry1ListCombo.append(ctryList[ctryIndex[0]])
                    ctry2ListCombo.append(ctryList[ctryIndex[1]])
                    EAListCombo.append(inParaCombos[ixCombo][0 + 2 * ctryIndex[0]])
                    sigmaAListCombo.append(inParaCombos[ixCombo][1 + 2 * ctryIndex[0]])
                    EBListCombo.append(inParaCombos[ixCombo][0 + 2 *ctryIndex[1]])
                    sigmaBListCombo.append(inParaCombos[ixCombo][1 + 2 * ctryIndex[1]])
                self.paraCombos.append(zip(ctry1ListCombo, ctry2ListCombo, EAListCombo, sigmaAListCombo, EBListCombo, sigmaBListCombo))
                ctry1List.extend(ctry1ListCombo)
                ctry2List.extend(ctry2ListCombo)
                EAList.extend(map(str, EAListCombo))
                EBList.extend(map(str, EBListCombo))
                sigmaAList.extend(map(str, sigmaAListCombo))
                sigmaBList.extend(map(str, sigmaBListCombo))



            variables = ['Ctry', 'Ctry', 'EA', 'EB', 'sigmaA', 'sigmaB']
            groups    = [ 0, 0, 0, 0, 0, 0 ]
            groupRestrs = [ 'diagnol', 'diagnol', 'diagnol', 'diagnol', 'diagnol', 'diagnol' ]
            writeVals = [ ", ".join(ctry1List), ", ".join(ctry2List), ", ".join(EAList), ", ".join(EBList), ", ".join(sigmaAList),", ".join(sigmaBList)]
            paraFiles = [ 'input/markovA.in', 'input/markovB.in', 'input/parameters.in', 'input/parameters.in', 'input/parameters.in', 'input/parameters.in' ]
            paraFileRegex = [ 'space-separated', 'space-separated', 'bar-separated', 'bar-separated' , 'bar-separated' , 'bar-separated'  ]
            #self.paraCombos = inParaCombos
            self.analyzeResults.tablePath = self.iterationFolder
            # variable list passed to analyzeOverviewTables
            self.variables = ['EA', 'sigmaA', 'EB', 'sigmaB']
            print 'Done setting up one4eachCtry. '




        # Write a para.loop file to generate grid jobs
        para_loop = self.writeParaLoop(variables = variables,
                                       groups = groups,
                                       groupRestrs = groupRestrs,
                                       vals = writeVals,
                                       desPath = os.path.join(self.iterationFolder, 'para.loopTmp'),
                                       paraFiles = paraFiles,
                                       paraFileRegex = paraFileRegex)

        paraLoop_fp.__init__(self, verbosity = self.verbosity)
        tasks = self.generateTaskList(para_loop, self.iterationFolder)
        ParallelTaskCollection.__init__(self, self.jobname, tasks)

    def target(self, inParaCombos):

        logger.debug('entering gParaSearchParallel.target')
        # Each line in the resulting table (overviewSimu) represents one paraCombo
        overviewTable = createOverviewTable(resultDir = self.iterationFolder, outFile = 'simulation.out', slUIPFile = 'slUIP.mat',
                                            exportFileName = 'overviewSimu', sortTable = False,
                                            logLevel = self.verbosity, logFile = os.path.join(self.pathToStageDir, 'createOverviewTable.log'))

        result = self.analyzeResults(tableIn = overviewTable, varsIn = self.variables, valsIn = self.paraCombos,
                             targetVar = 'normDev', logLevel = self.verbosity,
                             logFile = os.path.join(self.iterat
                                                    ionFolder, os.path.join(self.pathToStageDir, 'analyzeOverview.log')))
        #result = [ ele[0] for ele in inParaCombos ]
        logger.info('Computed target: Returning result to solver')
        self.iteration += 1
        return result


    def print_status(self, mins,means,vector,txt):
        print txt,mins, means, list(vector)

    def generateTaskList(self, para_loop, iterationFolder):
        # Fill the task list
        tasks = []
        for jobname, substs in self.process_para_file(para_loop):
            executable = os.path.basename(self.executable)
            # start the inputs dictionary with syntax: client_path: server_path
            inputs = { self.executable:executable }
            # make a "stage" directory where input files are collected on the client machine.
            #path_to_stage_dir = self.make_directory_path(os.path.join(iterationFolder, 'NAME'), jobname)
            path_to_stage_dir = os.path.join(iterationFolder, jobname)
            # input_dir is cwd/jobname (also referred to as "stage" dir.
            input_dir = path_to_stage_dir
            gc3libs.utils.mkdir(input_dir)
            prefix_len = len(input_dir) + 1
            # 1. files in the "initial" dir are copied verbatim
            # ugly work around to determine the correct base dir in case of one4all.
            baseDir = os.path.join(os.getcwd(), 'base')
            for (path, changes) in substs.iteritems():
                for (var, val, index, regex) in changes:
                    if var == 'Ctry':
                        baseDir += val
            if baseDir == 'base':
                baseDir = self.baseDir
            self.fillInputDir( baseDir, input_dir)
            # 2. apply substitutions to parameter files
            for (path, changes) in substs.iteritems():
                for (var, val, index, regex) in changes:
                    # new. make adjustments in the base dir itself.
                    update_parameter_in_file(os.path.join(input_dir, path),
                                             var, index, val, regex)
            # 3. build input file list
            for dirpath,dirnames,filenames in os.walk(input_dir):
                for filename in filenames:
                    # cut the leading part, which is == to path_to_stage_dir
                    relpath = dirpath[prefix_len:]
                    # ignore output directory contents in resubmission
                    if relpath.startswith('output'):
                        continue
                    remote_path = os.path.join(relpath, filename)
                    inputs[os.path.join(dirpath, filename)] = remote_path
            # all contents of the `output` directory are to be fetched
            outputs = { 'output/':'' }
            #kwargs = extra.copy()
            kwargs = self.extra_args
            kwargs['stdout'] = 'forwardPremiumOut.log'
            kwargs['join'] = True
            kwargs['output_dir'] = os.path.join(path_to_stage_dir, 'output')
            kwargs['requested_architecture'] = self.architecture
            kwargs['requested_cores'] = 1
            # hand over job to create
            tasks.append(GPremiumApplication('./' + executable, [], inputs, outputs, **kwargs))
        return tasks

class gParaSearchScript(SessionBasedScript, paraLoop_fp):
    """
    Read `.loop` files and execute the `forwardPremium` program accordingly.
    """

    def __init__(self):

        SessionBasedScript.__init__(
            self,
            version = '0.2',
            # only '.loop' files are considered as valid input
            input_filename_pattern = '*.loop',
        )



    def setup_options(self):
        self.add_param("-b", "--initial", metavar="DIR",
                       dest="initial",
                       help="Include directory contents in any job's input."
                       " Use this to specify the initial guess files.")
        self.add_param("-n", "--dry-run",
                       dest = 'dryrun', action="store_true", default = False,
                       help="Take the loop for a test run")
        self.add_param("-x", "--executable", metavar="PATH",
                       dest="executable", default=os.path.join(
                           os.getcwd(), "forwardPremiumOut"),
                       help="Path to the `forwardPremium` executable binary"
                       "(Default: %(default)s)")
        self.add_param("-X", "--architecture", metavar="ARCH",
                       dest="architecture", default=Run.Arch.X86_64,
                       help="Processor architecture required by the executable"
                       " (one of: 'i686' or 'x86_64', without quotes)")
        self.add_param("-NP", "--nPopulation", metavar="ARCH",
                       dest="nPopulation", default=5,
                       help="Population size for global optimizer")
        self.add_param("-xVars", "--xVars", metavar="ARCH",
                       dest="xVars", default = 'EA',
                       help="x variables over which to optimize")
        self.add_param("-xVarsDom", "--xVarsDom", metavar="ARCH",
                       dest="xVarsDom", default = '0.5 0.9',
                       help="Domain to sample x values from. Space separated list. ")
        self.add_param("-sv", "--solverVerb", metavar="ARCH",
                       dest="solverVerb", default = '0.5 0.9',
                       help="Separate verbosity level for the global optimizer ")
        self.add_param("-t", "--problemType", metavar="ARCH",
                       dest="problemType", default = 'one4eachPair',
                       help="Problem type for gParaSearch. Must be one of: one4eachPair, one4all, one4eachCtry. ")
        self.add_param("-e", "--pathEmpirical", metavar="PATH",
                       dest="pathEmpirical", default = '',
                       help="Path to empirical analysis folder")
        self.add_param("-i", "--itermax", metavar="ARCH", type = int,
                       dest="itermax", default = '50',
                       help="Maximum number of iterations of solver. ")
        self.add_param("-xC", "--xConvCrit", metavar="ARCH", type = float,
                       dest="xConvCrit", default = '1.e-8',
                       help="Convergence criteria for x variables. ")
        self.add_param("-yC", "--yConvCrit", metavar="ARCH", type = float,
                       dest="yConvCrit", default = '1.e-3',
                       help="Convergence criteria for y variables. ")
        self.add_param("-mP", "--makePlots", metavar="ARCH", type = bool,
                       dest="makePlots", default = True,
                       help="Generate population plots each iteration.  ")
        self.add_param("-oS", "--optStrategy", metavar="ARCH", type = int,
                       dest="optStrategy", default = '1',
                       help="Which differential evolution technique to use. ")
        self.add_param("-fW", "--fWeight", metavar="ARCH", type = float,
                       dest="fWeight", default = '0.85',
                       help="Weight of differential vector. ")
        self.add_param("-fC", "--fCritical", metavar="ARCH", type = float,
                       dest="fCritical", default = '1.0',
                       help="Fraction of new population to use.  ")
        self.add_param("-cL", "--countryList", metavar="ARCH", type = str,
                       dest="countryList", default = 'AU UK',
                       help="List of countries to analyze. ")
        self.add_param("-nN", "--norm", metavar="ARCH", type = str,
                       dest="norm", default = 2,
                       help="Which norm to apply for one4all and one4eachCtry")
        self.add_param("-iP", "--initialPop", metavar="ARCH", type = str,
                       dest="initialPop", default = "",
                       help="File name that contains a space separated list of initial pop. ")

    def parse_args(self):
        """
        Check validity and consistency of command-line options.
        """
        if not os.path.exists(self.params.executable):
            raise gc3libs.exceptions.InvalidUsage(
                "Path '%s' to the 'forwardPremium' executable does not exist;"
                " use the '-x' option to specify a valid one."
                % self.params.executable)
        if os.path.isdir(self.params.executable):
            self.params.executable = os.path.join(self.params.executable,
                                                  'forwardPremium')
        gc3libs.utils.check_file_access(self.params.executable, os.R_OK|os.X_OK,
                                gc3libs.exceptions.InvalidUsage)


    def new_tasks(self, extra):

        logger.info('entering new_tasks')

        # Adjust logging level for stream handler. Could be extracted into a general function.
        for handler in logger.handlers:
            if not isinstance(handler, logbook.FileHandler):
                logger.handlers.remove(handler)
        mySH = logbook.StreamHandler(stream = sys.stdout, level = self.params.solverVerb.upper(), format_string = '{record.message}', bubble = True)
        logger.handlers.append(mySH)

        baseDir = self.params.initial
        xVars    = self.params.xVars

        countryList = self.params.countryList.split()

        ctryIndices = getIndex(base = [len(countryList), len(countryList)], restr = 'lowerTr')
        for ctryIndex in ctryIndices:
            logger.info(countryList[ctryIndex[0]] + countryList[ctryIndex[1]])


        # Compute domain
        xVarsDom = self.params.xVarsDom.split()
        lowerBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 0], dtype = 'float64')
        upperBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 1], dtype = 'float64')
        domain = zip(lowerBds, upperBds)


        # Make problem type specific adjustments.
        if self.params.problemType == 'one4all':
            gdpTable = tableDict.fromTextFile(fileIn = os.path.join(self.params.pathEmpirical, 'outputInput/momentTable/Gdp/gdpMoments.csv'),
                                              delim = ',', width = 20)

            jobname = 'one4all'

            logger.info('%s' % self.params.norm)
            norm = self.params.norm
            try:
                norm = int(norm)
            except ValueError:
                if norm == "np.inf" or norm == "inf":
                    norm = np.inf
                else:
                    pass # in particular custom norms

            logger.info('using norm %s' % (norm))


            path_to_stage_dir = os.getcwd()
            executable = os.path.basename(self.params.executable)
            analyzeResults = anaOne4all(len(list(ctryIndices)), norm = norm)
            nlc            = nlcOne4all(gdpTable = gdpTable, ctryList = countryList, domain = domain, logFile = os.path.join(path_to_stage_dir, 'nlc.log'))
            combOverviews  = combineOverviews.combineOverviews(overviewSimuFile = 'eSigmaTable', tableName = 'ag_eSigmaTable', sortKeys = ['norm'])
            plot3dTable    = combineOverviews.plotTable(tablePath =os.path.join(path_to_stage_dir, 'ag_eSigmaTable'), savePath = os.path.join(path_to_stage_dir, 'scatter3d'))
            plot3dTable.columnNames = ['E', 'sigma', 'norm']
            for ctryIndex in ctryIndices:
                Ctry1 = countryList[ctryIndex[0]]
                Ctry2 = countryList[ctryIndex[1]]
                # Set Ctry information for this run.
                update_parameter_in_file(os.path.join(baseDir, 'input/markovA.in'), 'Ctry',
                                          0,  Ctry1,  'space-separated')
                update_parameter_in_file(os.path.join(baseDir, 'input/markovB.in'), 'Ctry',
                                          0,  Ctry2,  'space-separated')
                # Get the correct Ctry Paras into base dir.
                self.getCtryParas(baseDir, Ctry1, Ctry2)
                # Copy base dir
                ctryBaseDir = os.path.join(path_to_stage_dir, 'base' + Ctry1 + Ctry2)
                try:
                    shutil.copytree(baseDir, ctryBaseDir)
                except:
                    logger.info('%s already exists' % baseDir)

            kwargs = extra.copy()
            kwargs['output_dir'] = path_to_stage_dir

            # yield job
            yield (jobname, gParaSearchDriver,
                   [ self.params.executable, path_to_stage_dir, self.params.architecture,
                     baseDir, self.params.xVars, self.params.initialPop,
                     self.params.nPopulation, domain, self.params.solverVerb, self.params.problemType,
                     self.params.pathEmpirical, self.params.itermax, self.params.xConvCrit, self.params.yConvCrit,
                     self.params.makePlots, self.params.optStrategy, self.params.fWeight, self.params.fCritical, self.params.countryList,
                     analyzeResults, nlc, plot3dTable, combOverviews
                   ], kwargs)



        elif self.params.problemType == 'one4eachPair':
            for ctryIndex in ctryIndices:
                Ctry1 = countryList[ctryIndex[0]]
                Ctry2 = countryList[ctryIndex[1]]
                logger.info(Ctry1 + Ctry2)
                jobname = Ctry1 + '-' + Ctry2
                # set stage dir.
                path_to_stage_dir = self.make_directory_path(self.params.output, jobname)
                gc3libs.utils.mkdir(path_to_stage_dir)
                #path_to_stage_dir = os.path.join(iterationFolder, jobname)
                # Get moments table from empirical analysis
                gdpTable = tableDict.fromTextFile(fileIn = os.path.join(self.params.pathEmpirical, 'outputInput/momentTable/Gdp/gdpMoments.csv'),
                                                  delim = ',', width = 20)

                # Set Ctry information for this run.
                update_parameter_in_file(os.path.join(baseDir, 'input/markovA.in'), 'Ctry',
                                          0,  Ctry1,  'space-separated')
                update_parameter_in_file(os.path.join(baseDir, 'input/markovB.in'), 'Ctry',
                                          0,  Ctry2,  'space-separated')
                # Get the correct Ctry Paras into base dir.
                self.getCtryParas(baseDir, Ctry1, Ctry2)
                # Copy base dir
                ctryBaseDir = os.path.join(path_to_stage_dir, 'base')
                try:
                    shutil.copytree(baseDir, ctryBaseDir)
                except:
                    logger.info('%s already exists' % baseDir)
                EA = getParameter(fileIn = os.path.join(baseDir, 'input/parameters.in'), varIn = 'EA',
                                     regexIn = 'bar-separated')
                EB = getParameter(fileIn = os.path.join(baseDir, 'input/parameters.in'), varIn = 'EB',
                                     regexIn = 'bar-separated')
                sigmaA = getParameter(fileIn = os.path.join(baseDir, 'input/parameters.in'), varIn = 'sigmaA',
                                     regexIn = 'bar-separated')
                sigmaB = getParameter(fileIn = os.path.join(baseDir, 'input/parameters.in'), varIn = 'sigmaB',
                                     regexIn = 'bar-separated')
                # Pass ctry information to nlc
                analyzeResults = anaOne4eachPair
                nlc = nlcOne4eachPair(gdpTable = gdpTable, ctryPair = [Ctry1, Ctry2], domain = domain, logFile = os.path.join(path_to_stage_dir, 'nlc.log'))
                combOverviews = combineOverviews.combineOverviews(overviewSimuFile = 'overviewSimu', tableName = 'agTable', sortKeys = ['normDev'])
                plot3dTable    = combineOverviews.plotTable(tablePath =os.path.join(path_to_stage_dir, 'agTable'), savePath = os.path.join(path_to_stage_dir, 'scatter3d'))
                plot3dTable.columnNames = ['EA', 'sigmaA', 'normDev']

                executable = os.path.basename(self.params.executable)
                kwargs = extra.copy()
                kwargs['output_dir'] = path_to_stage_dir

                # yield job
                yield (jobname, gParaSearchDriver,
                       [ self.params.executable, path_to_stage_dir, self.params.architecture,
                         ctryBaseDir, self.params.xVars, self.params.initialPop,
                         self.params.nPopulation, domain, self.params.solverVerb, self.params.problemType,
                         self.params.pathEmpirical, self.params.itermax, self.params.xConvCrit, self.params.yConvCrit,
                         self.params.makePlots, self.params.optStrategy, self.params.fWeight, self.params.fCritical, self.params.countryList,
                         analyzeResults, nlc, plot3dTable, combOverviews
                       ], kwargs)

        elif self.params.problemType == 'one4eachCtry':
            gdpTable = tableDict.fromTextFile(fileIn = os.path.join(self.params.pathEmpirical, 'outputInput/momentTable/Gdp/gdpMoments.csv'),
                                              delim = ',', width = 20)

            if len(countryList) > len(self.params.xVars.split()) / 2 and len(xVars.split()) == 2:
                if self.params.xVars[-1:] != " ":
                    xVars = ( self.params.xVars + ' ' ) * len(countryList)
                    xVars = xVars[:-1]
                else:
                    xVars = self.params.xVars * len(countryList)
                if self.params.xVarsDom[-1:] != " ":
                    xVarsDom = ( self.params.xVarsDom + ' ' ) * len(countryList)
                    xVarsDom = xVarsDom[:-1].split()
                else:
                    xVarsDom = self.params.xVarsDom * len(countryList)
                    xVarsDom = xVarsDom.split()
            else:
                xVars = self.params.xVars
                xVarsDom = self.params.xVarsDom[:-1].split()
            jobname = 'one4eachCtry'

            lowerBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 0], dtype = 'float64')
            upperBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 1], dtype = 'float64')
            domain = zip(lowerBds, upperBds)

            norm = self.params.norm
            try:
                norm = int(norm)
            except ValueError:
                if norm == "np.inf" or norm == "inf":
                    norm = np.inf
                else:
                    pass # in particular custom norms

            logger.info('using norm %s' % (norm))

            path_to_stage_dir = os.getcwd()
            executable = os.path.basename(self.params.executable)
            analyzeResults = anaOne4eachCtry(countryList, len(list(ctryIndices)), norm = norm)
            nlc            = nlcOne4eachCtry(gdpTable = gdpTable, ctryList = countryList, domain = domain, logFile = os.path.join(path_to_stage_dir, 'nlc.log'))
            combOverviews  = combineOverviews.combineOverviews(overviewSimuFile = 'eSigmaTable', tableName = 'ag_eSigmaTable', sortKeys = ['norm'])
            deKenPrice.plotPopulation = plotPopOne4eachCtry(countryList)


##            # Set solver variables
##            nXvars = len(xVars.split())
##            deKenPrice.I_NP         = int(self.params.nPopulation)
##            deKenPrice.F_weight     = float(self.params.fWeight)
##            deKenPrice.F_CR         = float(self.params.fCritical)
##            deKenPrice.I_D          = int(nXvars)
##            deKenPrice.lowerBds     = np.array([ element[0] for element in domain ], dtype = 'float64')
##            deKenPrice.upperBds     = np.array([ element[1] for element in domain ], dtype = 'float64')
##            deKenPrice.I_itermax    = int(self.params.itermax)
##            deKenPrice.F_VTR        = float(self.params.yConvCrit)
##            deKenPrice.I_strategy   = int(self.params.optStrategy)
##            deKenPrice.I_plotting   = int(self.params.makePlots)
##            deKenPrice.xConvCrit    = float(self.params.xConvCrit)
##            deKenPrice.workingDir   = path_to_stage_dir
##            deKenPrice.verbosity    = self.params.solverVerb


            plot3dTable    = emptyFun
#            plot3dTable    = combineOverviews.plotTable(tablePath =os.path.join(path_to_stage_dir, 'ag_eSigmaTable'), savePath = os.path.join(path_to_stage_dir, 'scatter3d'))
#            plot3dTable.columnNames = ['E', 'sigma', 'norm']
            for ctryIndex in ctryIndices:
                Ctry1 = countryList[ctryIndex[0]]
                Ctry2 = countryList[ctryIndex[1]]
                # Set Ctry information for this run.
                update_parameter_in_file(os.path.join(baseDir, 'input/markovA.in'), 'Ctry',
                                          0,  Ctry1,  'space-separated')
                update_parameter_in_file(os.path.join(baseDir, 'input/markovB.in'), 'Ctry',
                                          0,  Ctry2,  'space-separated')
                # Get the correct Ctry Paras into base dir.
                self.getCtryParas(baseDir, Ctry1, Ctry2)
                # Copy base dir
                ctryBaseDir = os.path.join(path_to_stage_dir, 'base' + Ctry1 + Ctry2)
                try:
                    shutil.copytree(baseDir, ctryBaseDir)
                except:
                    logger.info('%s already exists' % baseDir)

            kwargs = extra.copy()
            kwargs['output_dir'] = path_to_stage_dir

            # yield job

            yield (jobname, gParaSearchDriver,
                   [ self.params.executable, path_to_stage_dir, self.params.architecture,
                     baseDir, xVars, self.params.initialPop,
                     self.params.nPopulation, domain, self.params.solverVerb, self.params.problemType,
                     self.params.pathEmpirical, self.params.itermax, self.params.xConvCrit, self.params.yConvCrit,
                     self.params.makePlots, self.params.optStrategy, self.params.fWeight, self.params.fCritical, self.params.countryList,
                     analyzeResults, nlc, plot3dTable, combOverviews
                   ], kwargs)

        # Set solver variables
        nXvars = len(xVars.split())
        deKenPrice.I_NP         = int(self.params.nPopulation)
        deKenPrice.F_weight     = float(self.params.fWeight)
        deKenPrice.F_CR         = float(self.params.fCritical)
        deKenPrice.I_D          = int(nXvars)
        if self.params.problemType == 'one4eachCtry':
            deKenPrice.lowerBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 0], dtype = 'float64')
            deKenPrice.upperBds = np.array([xVarsDom[i] for i in range(len(xVarsDom)) if i % 2 == 1], dtype = 'float64')
        else:
            deKenPrice.lowerBds     = np.array([ element[0] for element in domain ], dtype = 'float64')
            deKenPrice.upperBds     = np.array([ element[1] for element in domain ], dtype = 'float64')
        deKenPrice.I_itermax    = int(self.params.itermax)
        deKenPrice.F_VTR        = float(self.params.yConvCrit)
        deKenPrice.I_strategy   = int(self.params.optStrategy)
        deKenPrice.I_plotting   = int(self.params.makePlots)
        deKenPrice.xConvCrit    = float(self.params.xConvCrit)
        deKenPrice.workingDir   = path_to_stage_dir
        deKenPrice.verbosity    = self.params.solverVerb

        logger.info('done with new_tasks')


# run script

if __name__ == '__main__':
    logger.info('Starting: \n%s' % ' '.join(sys.argv))
    gParaSearchScript().run()
    logger.info('main done')



#! /usr/bin/env python
#
"""
This is the main entry point for command `gc3utils`:command: -- a
simple command-line frontend to distributed resources

This is a generic front-end code; actual implementation of commands
can be found in `gc3utils.commands`:mod:
"""
# Copyright (C) 2009-2012, 2019  University of Zurich. All rights reserved.
#
# Includes parts adapted from the ``bzr`` code, which is
# copyright (C) 2005, 2006, 2007, 2008, 2009 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


# stdlib imports
from __future__ import absolute_import, print_function, unicode_literals

import os
import os.path
import sys


def main():
    """
    Generic front-end function to invoke the commands in `gc3utils/commands.py`
    """

    # program name
    PROG = os.path.basename(sys.argv[0])

    # use docstrings for providing help to users,
    # so complain if they were removed by excessive optimization
    if __doc__ is None:
        sys.stderr.write(
            "%s does not support python -OO; aborting now.\n" % PROG)
        sys.exit(2)

    # ensure we run on a supported Python version
    NEED_VERS = (2, 4)
    try:
        version_info = sys.version_info
    except AttributeError:
        version_info = 1, 5  # 1.5 or older
    REINVOKE = "__GC3UTILS_REINVOKE"
    KNOWN_PYTHONS = ('python2.6', 'python2.5', 'python2.4')
    if version_info < NEED_VERS:
        if REINVOKE not in os.environ:
            # mutating os.environ doesn't work in old Pythons
            os.putenv(REINVOKE, "1")
            for python in KNOWN_PYTHONS:
                try:
                    os.execvp(python, [python] + sys.argv)
                except OSError:
                    pass
        sys.stderr.write("%s: error: cannot find a suitable python interpreter"
                         " (need %d.%d or later)" % (PROG, NEED_VERS))
        return 1
    if hasattr(os, "unsetenv"):
        os.unsetenv(REINVOKE)

    # ensure locale is set to "" (C, POSIX); otherwise parsing messages from
    # commands we invoke might fail because they speak, e.g., German.
    #
    if sys.platform == 'darwin':
        # jameinel says this hack is to force python to honor the LANG setting,
        # even on Darwin.  Otherwise it is apparently hardcoded to Mac-Roman,
        # which is incorrect for the normal Terminal.app which wants UTF-8.
        #
        # "It might be that I should be setting the "system locale"
        # somewhere else on the system, rather than setting
        # LANG=en_US.UTF-8 in .bashrc.  Switching to 'posix' and
        # setting LANG worked for me."
        #
        # So we can remove this if someone works out the right way to tell Mac
        # Python which encoding to use.  -- mbp 20080703
        sys.platform = 'posix'
        try:
            import locale
        finally:
            sys.platform = 'darwin'
    else:
        import locale

    try:
        locale.setlocale(locale.LC_ALL, '')
    except locale.Error as e:
        sys.stderr.write(
            '%s: WARNING: %s\n'
            '  could not set the application locale.\n'
            '  This might cause problems with some commands.\n'
            '  To investigate the issue, look at the output\n'
            '  of the locale(1p) tool available on POSIX systems.\n'
            % (PROG, e))

    if PROG == 'gc3utils':
        # the real command name is the first non-option argument
        for arg in sys.argv[1:]:
            if not arg.startswith('-'):
                if arg.startswith('g'):
                    PROG = arg
                else:
                    PROG = 'g' + arg
                del sys.argv[0]  # so that PROG == sys.argv[0]
                break

        # no command name found, print usage text and exit
        if PROG == 'gc3utils':
            sys.stderr.write("""Usage: gc3utils COMMAND [options]

Command `gc3utils` is a unified front-end to computing resources.
You can get more help on a specific sub-command by typing::

  gc3utils COMMAND --help

where command is one of these:
""")
            import gc3utils.commands
            for cmd in [sym[5:] for sym in dir(gc3utils.commands)
                        if sym.startswith("cmd_")]:
                sys.stderr.write('  ' + cmd + '\n')
            return 1

    # find command as function in the `commands.py` module
    PROG.replace('-', '_')
    import gc3utils.commands
    try:
        cmd = getattr(gc3utils.commands, 'cmd_' + PROG)
    except AttributeError:
        sys.stderr.write(
            "Cannot find command '%s' in gc3utils; aborting now.\n" % PROG)
        return 1
    rc = cmd().run()  # (*sys.argv[1:])
    return rc



#! /usr/bin/env python
#
#   gfittingaddm.py -- Front-end script for running MCMC with opneBUGS
#   function throguh R over a large table set.
#
#   Copyright (C) 2018, 2019  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Front-end script for submitting multiple `R` jobs.
It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gfittingaddm.py --help`` for program usage
instructions.

arguments for the 'main' R script to be executed:
@int:subject_number
@int:number_of_simulations_per_subject
@int:number_of_iteration_per_simulation
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2018-01-10:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '0.1.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gfittingaddm
    gfittingaddm.GfittingaddmScript().run()

import os
import sys
import time
import tempfile

import shutil

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file, positive_int, existing_directory
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask, StagedTaskCollection, ParallelTaskCollection

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(1500,MB)
DEFAULT_WALLTIME = Duration(300,hours)
DEFAULT_RESULTS = "Results"
DEFAULT_SIMULATIONS=100
DEFAULT_ITERATIONS=150

## custom application class
class GfittingaddmApplication(Application):
    """
    Custom class to wrap the execution of the R scripts passed in src_dir.
    """
    application_name = 'gfittingaddm'

    def __init__(self, subject_number,  rscript_folder, main_function, n_simulations, n_iterations, **extra_args):

        inputs = {}
        outputs = {}
        executables = []

        if rscript_folder:
            inputs[rscript_folder] = "./"

        outputs[DEFAULT_RESULTS] = DEFAULT_RESULTS

        arguments = "Rscript --vanilla {0}.R {1} {2} {3}".format(main_function,
                                                                 subject_number,
                                                                 n_simulations,
                                                                 n_iterations)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = outputs,
            stdout = 'gfittingaddm.log',
            executables = executables,
            join=True,
            **extra_args)

class GfittingaddmScript(SessionBasedScript):
    """
    The ``gfittingaddm`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gfittingaddm`` will delay submission of
    newly-created jobs so that this limit is never exceeded.

    Once the processing of all chunked files has been completed, ``gfittingaddm``
    aggregates them into a single larger output file located in
    'self.params.output'.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__,
            application = GfittingaddmApplication,
            stats_only_for = GfittingaddmApplication,
            )

    def setup_args(self):
        self.add_param('number_of_subjects', type=positive_int,
                       help="Number of subjects to be analysed")

    def setup_options(self):
        self.add_param("-M", "--Rscripts", metavar="[PATH]",
                       type=existing_directory,
                       dest="Rscripts", default=None,
                       help="Location of R Main scripts. Default: '%(default)s'.")

        self.add_param("-F", "--main_function", metavar="[PATH]",
                       dest="main_function", default="runme",
                       help="R script main function to be invoked. Default: '%(default)s'.")

        self.add_param("-R", "--repeat", metavar="[INT]",
                       type=positive_int,
                       dest="repeat", default=1,
                       help="Repeat each subject simulation. Default: '%(default)s'.")

        self.add_param("-S", "--simulations", metavar="[INT]",
                       type=positive_int,
                       dest="simulations", default=DEFAULT_SIMULATIONS,
                       help="Number of simulations for each individual subject." \
                       " Default: '%(default)s'.")

        self.add_param("-I", "--iterations", metavar="[INT]",
                       type=positive_int,
                       dest="iterations", default=DEFAULT_ITERATIONS,
                       help="Number of iterations for each individual simulation." \
                       " Default: '%(default)s'.")

        self.add_param("-F", "--follow",
                       dest="follow",
                       action="store_true",
                       default=False,
                       help="Periodically fetch job's output folder and copy locally." \
                       " Default: '%(default)s'.")

    def before_main_loop(self):
        # XXX: should this be done with `make_controller` instead?
        self._controller.retrieve_running = self.params.follow
        self._controller.retrieve_overwrites = self.params.follow
        self._controller.retrieve_changed_only = self.params.follow

    def new_tasks(self, extra):
        """
        Chunk initial input file
        For each chunked fule, generate a new GfittingaddmTask
        """
        tasks = []

        for subject_number in range(1,self.params.number_of_subjects+1):
            for rep in range(1,self.params.repeat+1):
                extra_args = extra.copy()
                extra_args['jobname'] = "subject{0}-rep{1}".format(subject_number,
                                                                   rep)
                gc3libs.log.info("Creating task for subject number {0}".format(subject_number))

                tasks.append(GfittingaddmApplication(subject_number,
                                                     self.params.Rscripts,
                                                     self.params.main_function,
                                                     self.params.simulations,
                                                     self.params.iterations,
                                                     **extra_args))

        return tasks



#! /usr/bin/env python
#
#   gqhg.py -- Front-end script for running the QHGMain binary
#   over a list of intput population files .qdf
#
#   Copyright (C) 2016, 2017  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gqhg.py --help`` for program usage
instructions.
"""

from __future__ import absolute_import, print_function

# summary of user-visible changes
__changelog__ = """
  2016-07-01:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'
__version__ = '1.0.0'

# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gqhg
    gqhg.GqhgScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
import random

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(3000,MB)

DEFAULT_REMOTE_INPUT_FOLDER="./"
DEFAULT_REMOTE_OUTPUT_FOLDER="./output"
QDF_PATTERN = "qdf"
DEFAULT_ITERATIONS=10000
DEFAULT_RANDOM_LIMIT=1000000
DEFAULT_STEPS=1000
DEFAULT_EVENTS='write|geo|climate|veg@[1000],write|pop:Sapiens_ooa@[3000]+10000,write|stats@10000'
DEFAULT_QHMAIN_COMMAND="QHGMain --output-prefix=ooa "

## custom application class
class GqhgApplication(Application):
    """
    """
    application_name = 'gqhgi'

    def __init__(self, qdf_file, seed, steps, **extra_args):

        inputs = dict()
        outputs = dict()

        self.seed = seed
        output_logfile = "ooa_%d_08i" % seed
        qdf_filename = os.path.basename(qdf_file)

        inputs[qdf_file] = qdf_filename

        if 'grid' in extra_args:
            grid_filename = os.path.basename(extra_args['grid'])
            inputs[extra_args['grid']] = grid_filename

        for input_file in extra_args['event_files']:
            inputs[input_file]=os.path.basename(input_file)
            extra_args['event'] = extra_args['event'].replace(input_file,os.path.basename(input_file))

        arguments = DEFAULT_QHMAIN_COMMAND
        arguments += " --grid=%s " % grid_filename
        arguments += " --output-dir=%s " % DEFAULT_REMOTE_OUTPUT_FOLDER
        arguments += " --pops=%s " % qdf_filename
        arguments += " --shuffle=%d " %  seed
        arguments += " -n %d " % steps
        arguments += " --events=%s " % extra_args['event']
        arguments += " --log-file=%s.log > %s.out " % (output_logfile, output_logfile)

        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = [DEFAULT_REMOTE_OUTPUT_FOLDER, output_logfile+".log",
                       output_logfile+".out"],
            stdout = 'gqhg.log',
            join=True,
            **extra_args)

class GqhgScript(SessionBasedScript):
    """

    The ``gqhg`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.

    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gqhg`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GqhgApplication,
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GqhgApplication,
            )

    def setup_options(self):

        self.add_param("-G", "--grid", metavar="[PATH]",
                       dest="grid", default=None,
                       help="Location of the Grid file")

        self.add_param("-I", "--iterations",
                       dest="iterations", default=DEFAULT_ITERATIONS,
                       help="Repeat simulation on same .qdf file. "
                       "Default: %(default)s.")

        self.add_param("-n", "--steps",
                       dest="steps", default=DEFAULT_STEPS,
                       help="Number of steps within single simulation "
                       "Default: %(default)s.")

        self.add_param("-M", "--master",
                       dest="master", default=None,
                       help="Use alternative binanry file. "
                       "Note: binary has to be statically compiled")

        self.add_param("-E", "--events",
                       dest="events", default=DEFAULT_EVENTS,
                       help="Use specific `events` directive. " \
                       "assumes: "
                       "-- each event file has suffix .qdf "
                       "-- 'env' type events only (e.g. not 'write')"
                       "-- each event is separated by a comma"
                       "-- everything between : and .qdf is the filename"
                       "Default: %(default)s.")

    def setup_args(self):

        self.add_param('input_data', type=str,
                       help="Root localtion of input data. "
                       "Note: walk through input folder and select "
                       " all files ending with .qdf")

    def parse_args(self):
        """
        Check valid input files in input folder.
        For each valid input file execute reference binary and repear
        `self.params.iterantions` time.
        """

        try:

            assert os.path.isdir(self.params.input_data), \
                "Input folder '%s' not found" % self.params.input_data

            if self.params.grid:
                assert os.path.isfile(self.params.grid), \
                    "Grid file %s not found" % self.params.grid

            self.qdf_files = [ os.path.abspath(os.path.join(self.params.input_data,qdf))
                               for qdf in os.listdir(self.params.input_data)
                               if qdf.endswith(QDF_PATTERN) ]
            assert len(self.qdf_files) > 0, \
                "No QDF files found in input folder %s" % self.params.input_data

        except AssertionError as ex:
            raise ValueError(ex.message)

    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GqhgApplication
        """
        tasks = []

        list_input_qdf = parse_event_string(self.params.events)

        for qdf_file in self.qdf_files:

            for iteration in range(1,int(self.params.iterations)+1):

                seed = random.randint(1,DEFAULT_RANDOM_LIMIT)
                # extract root folder name to be used as jobname
                extra_args = extra.copy()
                jobname = "%s_%d" % (os.path.basename(qdf_file),iteration)
                extra_args['jobname'] = jobname

                extra_args['output_dir'] = self.params.output
                extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', jobname)
                extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', jobname)

                if self.params.grid:
                    extra_args['grid'] = os.path.abspath(self.params.grid)

                extra_args['event_files'] = list_input_qdf
                extra_args['event'] = self.params.events

                tasks.append(GqhgApplication(
                    qdf_file,
                    seed,
                    int(self.params.steps),
                    **extra_args))

        return tasks

def parse_event_string(eventstring):
    """Return a list of input event files given the command line event string."""
    """assumes: -- each event file has suffix .qdf
                -- 'env' type events only (e.g. not 'write')
                -- each event is separated by a comma
                -- events string begins with '-events='
                -- everything between : and .qdf is the filename
   """
    import string
    filesuffix = ".qdf"
    eventfilelist = []
    events = eventstring.split(",")
    for thisevent in events:
        if "=" in thisevent:   #   (e.g. '-events='.......)
            thisevent = thisevent.split("=")[1]
        first3letters = ""
        iletter = 0
        for thischar in thisevent: #   check 1st 3 letters for event 'tag' of env
            if thischar in string.ascii_letters:
                first3letters += thischar
                iletter += 1
            if iletter == 3:
                break
        if first3letters[:3] != "env":
            continue   # not the event we are looking for

        # filename starts after the ":"
        stringtmp = thisevent.split(":")[1].strip() # strip any space before filename
        stringtmp = stringtmp.split(filesuffix)[0]
        stringtmp += filesuffix
        eventfilelist.append(stringtmp)
    return eventfilelist



#! /usr/bin/env python

"""
Implementation of task collections.

Tasks can be grouped into collections, which are tasks themselves,
therefore can be controlled (started/stopped/cancelled) like a single
whole.  Collection classes provided in this module implement the basic
patterns of job group execution; they can be combined to form more
complex workflows.  Hook methods are provided so that derived classes
can implement problem-specific job control policies.
"""
# Copyright (C) 2009-2019   University of Zurich. All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import, print_function, unicode_literals
from builtins import str
from builtins import range
from builtins import object
__docformat__ = 'reStructuredText'

import itertools
import os

from collections import defaultdict
from gc3libs.compat.toposort import toposort

from gc3libs import Run, Task
import gc3libs.exceptions
import gc3libs.utils


class TaskCollection(Task):

    """
    Base class for all task collections. A "task collection" is a
    group of tasks, that can be managed collectively as a single one.

    A task collection implements the same interface as the `Task`
    class, so you can use a `TaskCollection` everywhere a `Task` is
    required.  A task collection has a `state` attribute, which is an
    instance of `gc3libs.Run.State`; each concrete collection class
    decides how to deduce a collective state based on the individual
    task states.
    """

    # a generic `TaskCollection` produces no output -- change in
    # subclasses or specific instance if needed
    would_output = False

    def __init__(self, tasks=None, **extra_args):
        if tasks is None:
            self.tasks = []
        else:
            self.tasks = tasks
        Task.__init__(self, **extra_args)

    def iter_workflow(self):
        """
        Returns an iterator that will traverse the whole tree of tasks.
        """
        return itertools.chain(
            # this task collection
            itertools.repeat(self, 1),
            # iterator over non-collection subtasks
            (task for task in self.tasks
                if not isinstance(task, TaskCollection)),
            # recurse into collection subtasks
            *(coll.iter_workflow() for coll in self.tasks
                if isinstance(coll, TaskCollection)))

    def iter_tasks(self):
        """
        Iterate over non-collection tasks enclosed in this collection.
        """
        return itertools.chain(
            # this task collection
            itertools.repeat(self, 1),
            # iterator over non-collection subtasks
            (task for task in self.tasks),
        )

    @property
    def changed(self):
        """
        Evaluates to `True` if this task or any of its subtasks has been
        modified and should be saved to persistent storage.
        """
        if self._changed:
            return True
        for task in self.tasks:
            if '_changed' in task and task._changed:
                return True
        return False

    @changed.setter
    def changed(self, value):
        self._changed = value

    # manipulate the "controller" interface used to control the associated task
    def attach(self, controller):
        """
        Use the given Controller interface for operations on the job
        associated with this task.
        """
        raise NotImplementedError(
            "Called abstract method TaskCollection.attach() -"
            " this should be overridden in derived classes.")

    def detach(self):
        for task in self.tasks:
            task.detach()
        Task.detach(self)

    def add(self, task):
        """
        Add a task to the collection.
        """
        raise NotImplementedError(
            "Called abstract method TaskCollection.add() -"
            " this should be overridden in derived classes.")

    def remove(self, task):
        """
        Remove a task from the collection.
        """
        self.tasks.remove(task)
        task.detach()

    # task execution manipulation -- these methods should be overriden
    # in derived classes, to implement the desired policy.

    def submit(self, resubmit=False, targets=None, **extra_args):
        raise NotImplementedError(
            "Called abstract method TaskCollection.submit() -"
            " this should be overridden in derived classes.")

    def update_state(self, **extra_args):
        """
        Update the running state of all managed tasks.
        """
        for task in self.tasks:
            if task.execution.state not in [Run.State.NEW, Run.State.TERMINATED]:
                gc3libs.log.debug(
                    "Updating state of task %s in collection %s ...",
                    task, self)
                task.update_state(**extra_args)

    def kill(self, **extra_args):
        # XXX: provide default implementation that kills all jobs?
        raise NotImplementedError(
            "Called abstract method TaskCollection.kill() -"
            " this should be overridden in derived classes.")

    def _get_download_dir(self, download_dir):
        """
        Return the base directory path where to download this Task
        collection's output files.

        Works just like method `Task._get_download_dir`, the only
        difference being that here we cannot return ``None`` if this
        Task is not expected to produce any output, as the value will
        be used as a base path for all dependent tasks.
        """
        # determine download dir
        if download_dir is not None:
            return download_dir
        else:
            try:
                return self.output_dir
            except AttributeError:
                cwd = os.getcwd()
                gc3libs.log.warning(
                    "`TaskCollection._get_download_dir()` called"
                    " with no explicit download directory,"
                    " but object '%s' (%s) has no `output_dir`"
                    " attribute set either. Using `%s` as collection's"
                    " base output directory."
                    % (self, type(self), cwd))
                return cwd

    def fetch_output(self, output_dir=None,
                     overwrite=False, changed_only=True, **extra_args):
        # if `output_dir` is not None, it is interpreted as the base
        # directory where to download files; each task will get its
        # own subdir based on its `.persistent_id`
        coll_output_dir = self._get_download_dir(output_dir)
        assert coll_output_dir is not None, \
            ("Unknown collection output directory!"
             " Task collection '%s' was not initialized with"
             " an explicit `output_dir=...` setting,"
             " but then `fetch_output()` was called without any"
             " explicit `output_dir=...` argument."
             % (self,))
        for task in self.tasks:
            if task.execution.state != Run.State.TERMINATING:
                continue
            if 'output_dir' in task:
                task_output_dir = task.output_dir
            else:
                task_output_dir = task.persistent_id
            # NOTE: uses a feature from `os.path.join`: if the second
            # path is absolute, the first path is discarded and the
            # second one is returned unchanged
            task_output_dir = os.path.join(coll_output_dir, task_output_dir)
            self._controller.fetch_output(
                task,
                task_output_dir,
                overwrite,
                changed_only,
                **extra_args)
        # NOTE: `all()` of an empty list comprehension is ``True``
        all_tasks_terminated = all(
            task.execution.state == Run.State.TERMINATED
            for task in self.tasks
        )
        if all_tasks_terminated:
            self.execution.state = Run.State.TERMINATED
        return coll_output_dir

    def free(self):
        """
        This method just asks the Engine to free the contained tasks.
        """
        if self._attached:
            for task in self.tasks:
                self._controller.free(task)

    def peek(self, what, offset=0, size=None, **extra_args):
        """
        Raise a `gc3libs.exceptions.InvalidOperation` error, as there
        is no meaningful semantics that can be defined for `peek` into
        a generic collection of tasks.
        """
        # is there any sensible semantic here?
        raise gc3libs.exceptions.InvalidOperation(
            "Cannot `peek()` on a task collection.")


    def stats(self, only=None):
        """
        Return a dictionary mapping each state name into the count of
        tasks in that state. In addition, the following keys are defined:

        * `ok`:  count of TERMINATED tasks with return code 0

        * `failed`: count of TERMINATED tasks with nonzero return code

        * `total`: count of managed tasks, whatever their state

        If the optional argument `only` is not None, tasks whose
        class is not contained in `only` are ignored.

        :param tuple only: Restrict counting to tasks of these classes.

        """
        result = defaultdict(int)
        for task in self.tasks:
            if only and not isinstance(task, only):
                continue
            state = task.execution.state
            result[state] += 1
            if state == Run.State.TERMINATED:
                if task.execution.returncode == 0:
                    result['ok'] += 1
                else:
                    result['failed'] += 1
        if only:
            result['total'] = len([task for task in self.tasks
                                   if isinstance(task, only)])
        else:
            result['total'] = len(self.tasks)
        return result

    def terminated(self):
        """
        Called when the job state transitions to `TERMINATED`, i.e.,
        the job has finished execution (with whatever exit status, see
        `returncode`) and the final output has been retrieved.

        Default implementation for `TaskCollection` is to set the
        exitcode to the maximum of the exit codes of its tasks,
        or ``None`` if no task has a numeric exit code.

        If no tasks were run, the exitcode is set to 0.
        """
        # Use a numeric argument instead of ``None``, so max() below
        # does not throw an error on Python 3.x because ``None`` and
        # integer numbers are not comparable.  Since the exit code is
        # limited to (signed) 8 bits, any number larger than that will do.
        NONE = -1000
        if self.tasks:
            exitcode = max(
                (task.execution._exitcode
                 if task.execution._exitcode is not None
                 else NONE)
                for task in self.tasks
            )
            if exitcode == NONE:
                exitcode = None
            self.execution._exitcode = exitcode
        else:
            # a sequence with no tasks terminates successfully
            self.execution._exitcode = 0
        gc3libs.log.debug("%s: exit code set to %s", self, self.execution.exitcode)


class SequentialTaskCollection(TaskCollection):

    """
    A `SequentialTaskCollection` runs its tasks one at a time.

    After a task has completed, the `next` method is called with the
    index of the finished task in the `self.tasks` list; the return
    value of the `next` method is then made the collection
    `execution.state`.  If the returned state is `RUNNING`, then the
    subsequent task is started, otherwise no action is performed.

    The default `next` implementation just runs the tasks in the order
    they were given to the constructor, and sets the state to
    `TERMINATED` when all tasks have been run.
    """

    def __init__(self, tasks, **extra_args):
        # XXX: check that `tasks` is a sequence type
        TaskCollection.__init__(self, tasks, **extra_args)
        self._current_task = None

    def add(self, task):
        task.detach()
        self.tasks.append(task)

    def attach(self, controller):
        """
        Use the given Controller interface for operations on the job
        associated with this task.
        """
        if self._current_task is not None:
            self.tasks[self._current_task].attach(controller)
        Task.attach(self, controller)

    def kill(self, **extra_args):
        """
        Stop execution of this sequence.  Kill currently-running task
        (if any), then set collection state to TERMINATED.
        """
        if self._current_task is not None:
            self.tasks[self._current_task].kill(**extra_args)
            for i in range(self._current_task+1, len(self.tasks)):
                self.tasks[i].execution.state = Run.State.TERMINATED
                self.tasks[i].execution.returncode = (Run.Signals.Cancelled, -1)
        self.execution.state = Run.State.TERMINATED
        self.execution.returncode = (Run.Signals.Cancelled, -1)
        self.changed = True

    def next(self, done):
        """
        Return collection state or task to run after step number `done` is terminated.

        This method is called when a task is finished; the `done`
        argument contains the index number of the just-finished task
        into the `self.tasks` list.  In other words, the task that
        just completed is available as `self.tasks[done]`.

        The return value from `next` can be either a task state (i.e.,
        an instance of `Run.State`), or a valid index number for
        `self.tasks`. In the first case:

        - if the return value is `Run.State.TERMINATED`,
          then no other jobs will be run;
        - otherwise, the return value is assigned to `execution.state`
          and the next job in the `self.tasks` list is executed.

        If instead the return value is a (nonnegative) number, then
        tasks in the sequence will be re-run starting from that index.

        The default implementation runs tasks in the order they were
        given to the constructor, and sets the state to TERMINATED
        when all tasks have been run.  This method can (and should) be
        overridden in derived classes to implement policies for serial
        job execution.
        """
        if done == len(self.tasks) - 1:
            return Run.State.TERMINATED
        else:
            return Run.State.RUNNING


    def progress(self):
        assert self._attached
        task = self.stage()
        if task is not None:
            task.progress()
        super(SequentialTaskCollection, self).progress()


    def redo(self, from_stage=0, *args, **kwargs):
        """
        Rewind the sequence to a given stage and reset its state to ``NEW``.

        In addition, when called with argument `from_stage` set to the
        total number of tasks in the collection, will try *continuing*
        the sequence by (ultimately) calling `self.next()` to get a
        new task.
        """
        nr_tasks = len(self.tasks)
        if nr_tasks > 0:
            assert from_stage <= nr_tasks, (
                "Cannot redo {0} from stage {1}:"
                " only {2} stages in task list."
                .format(self, from_stage, nr_tasks)
            )
            if from_stage < nr_tasks:
                self._current_task = from_stage
                task = self.stage()
                if task is not None:
                    task.redo(*args, **kwargs)
                # All other tasks should be put in NEW again
                for i in range(from_stage+1, nr_tasks):
                    self.tasks[i].redo(*args, **kwargs)
                super(SequentialTaskCollection, self).redo(*args, **kwargs)
            else:   # it's *continue* not *redo*
                assert self.tasks[-1].execution.state == Run.State.TERMINATED
                self.execution.state = Run.State.RUNNING
                self._current_task = nr_tasks - 1
                self.update_state()  # calls `next()`


    def submit(self, resubmit=False, targets=None, **extra_args):
        """
        Start the current task in the collection.
        """
        if self.tasks:
            if self._current_task is None:
                self._current_task = 0
            task = self.tasks[self._current_task]
            task.attach(self._controller)
            task.submit(resubmit, targets, **extra_args)
            if task.execution.state == Run.State.NEW:
                # submission failed, state unchanged
                self.execution.state = Run.State.NEW
            elif task.execution.state == Run.State.SUBMITTED:
                self.execution.state = Run.State.SUBMITTED
            else:
                self.execution.state = Run.State.RUNNING
        else:
            # no tasks to run, sequence is already finished
            self.execution.state = Run.State.TERMINATED
        self.changed = True
        return self.execution.state


    def stage(self):
        """
        Return the `Task` that is currently executing, or ``None``
        (if finished or not yet started).
        """
        if self._current_task is None:
            return None
        else:
            return self.tasks[self._current_task]


    def update_state(self, **extra_args):
        """
        Update state of the collection, based on the jobs' statuses.
        """
        task = self.stage()
        if task is None:
            # state is either NEW or TERMINATED, no update
            assert self.execution.state in [
                Run.State.NEW, Run.State.TERMINATED]
            return self.execution.state

        # update state of current task
        if task.execution.state not in [Run.State.NEW, Run.State.TERMINATED]:
            task.update_state(**extra_args)
        gc3libs.log.debug("Task `%s` in state %s", task, task.execution.state)

        # now set state based on the state of current task:
        #
        # 1. first task ever gets special treatment
        if (self._current_task == 0
            and task.execution.state in [
                Run.State.NEW,
                Run.State.SUBMITTED,
            ]):
            # avoid state flapping back to NEW if it's already SUBMITTED
            if self.execution.state == Run.State.NEW:
                self.execution.state = task.execution.state
            return self.execution.state

        # 2. if current task is terminated, advance to next one
        if (task.execution.state == Run.State.TERMINATED):
            nxt = self.next(self._current_task)
            if nxt in Run.State:
                self.execution.state = nxt
                collection_state_already_set = True
                if self.execution.state not in [
                        Run.State.STOPPED,
                        Run.State.TERMINATED,
                        Run.State.TERMINATING,
                ]:
                    self._current_task += 1
            else:
                # `nxt` must be a valid index into `self.tasks`
                assert 0 <= nxt < len(self.tasks)
                self._current_task = nxt
                collection_state_already_set = False
            # submit next task, unless we're TERMINATED or STOPPED
            if self.execution.state not in [
                    Run.State.STOPPED,
                    Run.State.TERMINATED,
                    Run.State.TERMINATING,
            ]:
                next_task = self.tasks[self._current_task]
                next_task.attach(self._controller)
                resubmit_task = (next_task.execution.state != Run.State.NEW)
                next_task.submit(resubmit=resubmit_task)
                if not collection_state_already_set:
                    self.execution.state = Run.State.RUNNING
                self.changed = True
            return self.execution.state

        # 3. if task stopped, stop the sequence too
        if (task.execution.state == Run.State.STOPPED):
            self.execution.state = Run.State.STOPPED
            return self.execution.state

        # 4. if task is running or terminating, keep on running
        if task.execution.state in [
                Run.State.NEW,
                Run.State.SUBMITTED,
                Run.State.RUNNING,
                Run.State.TERMINATING,
                Run.State.UNKNOWN,
        ]:
            self.execution.state = Run.State.RUNNING
            return self.execution.state

        # 5. this shouldn't happen!
        raise gc3libs.exceptions.InternalError(
            "Unhandled task state `{0}`"
            " in SequentialTaskCollection.update_state()"
            .format(task.execution.state))


class _OnError(object):
    """
    Mix-in class to make a `SequentialTaskCollection`:class: instance
    turn to a specific state as soon as one of the tasks fail.

    The final state is set by copying the `_on_error_state` attribute;
    of course it only makes sense that it is a state that stops
    further updates from GC3Pie -- like ``STOPPED`` (see
    `StopOnError`:class:) or ``TERMINATED`` (see
    `AbortOnError`:class:).

    A second effect of mixing this class in is that the
    `self.execution.returncode` attribute of the task collection
    instance mirrors the return code of the last finished task.

    .. seealso:: `StopOnError`:class:, `AbortOnError`:class:
    """

    # override in subclasses!
    _on_error_state = Run.State.UNKNOWN

    def complete(self):
        """
        Return ``True`` if the last executed task is the final task of the sequence.

        This is used in the `_OnError.next`:meth: to determine whether
        the last executed task is actually the last task of the
        sequence (hence the task collection is TERMINATED even if it
        failed), or more tasks could be added (hence the final state
        is set by `_on_error_state`).

        The default implementation always returns ``True``, which
        ensures that the mix-in classes `AbortOnError`:class: and
        `StopOnError`:class: work out-of-the-box with
        `StagedTaskCollection`:class: and
        `DependentTaskCollection`:task:.

        .. versionadded:: 2.5
        """
        return True

    def next(self, done):
        """
        Return collection state after step number `done` is terminated.

        Note that the task collection state should be set to
        ``TERMINATED`` after the last task has been executed,
        regardless of whether it failed or not.  Now, there are two
        distinct contexts where `next` could be applied in a
        `SequentialTaskCollection` instance:

        1. After the whole sequence has been built, i.e., when
           `self.tasks[-1]` is actually the last task that should be
           executed.  This is, e.g., the case with the stock
           `StagedTaskCollection`:class: and
           `DependentTaskCollection`:class: instances.  In this case,
           the `next` method can confidently mark the task collection
           as ``TERMINATED`` when the last task has been executed.

        2. While the sequence is still being built: in this case, a
           failed job should only set the task collection to
           `_on_error_state` -- as there might be further tasks coming
           down the road.

        Method `complete`:meth: is available to tell `next` which of
        these cases applies: if ``self.complete()`` returns ``True``
        then this class' implementation of `next` takes the behavior
        described in case 1. above; if `self.complete()` is instead
        ``False`` then the logic of case 2. above is applied instead.

        See `GitHub issue #512` for a lengthier discussion.

        .. seealso: :meth:`SequentialTaskCollection.next`, `GitHub issue #512`_

        .. _`GitHub issue #512`: https://github.com/uzh/gc3pie/issues/512
        """
        self.execution.returncode = self.tasks[done].execution.returncode
        if self.complete() and done == len(self.tasks)-1:
            return Run.State.TERMINATED
        else:
            if self.execution.returncode != 0:
                return self._on_error_state
            else:
                return Run.State.RUNNING


class AbortOnError(_OnError):
    """
    Mix-in class to make a `SequentialTaskCollection`:class: turn to
    ``TERMINATED`` state as soon as one of the tasks fail.

    A second effect of mixing this class in is that the
    `self.execution.returncode` mirrors the return code of the last
    finished task.


    .. note::

      For the mix-in to take effect, this class should be listed
      *before* the base task collection class, e.g.::

        # this works
        class MyTaskCollection(AbortOnError, SequentialTaskCollection):
          pass

        # this *does not* work
        class MyOtherTaskCollection(SequentialTaskCollection, AbortOnError):
          pass

    See :meth:`SequentialTaskCollection.next` and `GitHub issue #512`_
    for some caveats on applying this to dynamically-built task
    collections.

    .. _`GitHub issue #512`: https://github.com/uzh/gc3pie/issues/512
    """
    _on_error_state = Run.State.TERMINATED


class StopOnError(_OnError):
    """
    Mix-in class to make a `SequentialTaskCollection`:class: turn to
    ``STOPPED`` state as soon as one of the tasks fail.

    A second effect of mixing this class in is that the
    `self.execution.returncode` mirrors the return code of the last
    finished task.

    .. note::

      For the mix-in to take effect, this class should be listed
      *before* the base task collection class, e.g.::

        # this works
        class MyTaskCollection(StopOnError, SequentialTaskCollection):
          pass

        # this *does not* work
        class MyOtherTaskCollection(SequentialTaskCollection, StopOnError):
          pass

    See :meth:`SequentialTaskCollection.next` and `GitHub issue #512`_
    for some caveats on applying this to dynamically-built task
    collections.

    .. _`GitHub issue #512`: https://github.com/uzh/gc3pie/issues/512
    """
    _on_error_state = Run.State.STOPPED


class StagedTaskCollection(SequentialTaskCollection):

    """
    Simplified interface for creating a sequence of Tasks.
    This can be used when the number of Tasks to run is
    fixed and known at program writing time.

    A `StagedTaskCollection` subclass should define methods `stage0`,
    `stage1`, ... up to `stageN` (for some arbitrary value of N positive
    integer).  Each of these `stageN` must return a `Task`:class:
    instance; the task returned by the `stage0` method will be executed
    first, followed by the task returned by `stage1`, and so on.
    The sequence stops at the first N such that `stageN` is not defined.

    The exit status of the whole sequence is the exit status of the
    last `Task` instance run.  However, if any of the `stageN` methods
    returns an integer value instead of a `Task` instance, then the
    sequence stops and that number is used as the sequence exit
    code.

    """

    def __init__(self, **extra_args):
        try:
            first_stage = self.stage0()
            if isinstance(first_stage, Task):
                # init parent class with the initial task
                SequentialTaskCollection.__init__(
                    self, [first_stage], **extra_args)
            elif isinstance(first_stage, (int, int, tuple)):
                # init parent class with no tasks,
                # and immediately set the exitcode
                SequentialTaskCollection.__init__(self, [], **extra_args)
                self.execution.returncode = first_stage
                self.execution.state = Run.State.TERMINATED
            else:
                raise AssertionError(
                    "Invalid return value from method `stage0()` of"
                    " `StagedTaskCollection` object %r:"
                    " must return `Task` instance or integer exit code" % self)
        except AttributeError as ex:
            raise AssertionError(
                "Invalid `StagedTaskCollection` instance %r: %s"
                % (self, str(ex)))

    def next(self, done):
        # get next stage (1); if none exists, log it and exit
        try:
            next_stage_fn = getattr(self, "stage%d" % (done + 1))
        except AttributeError:
            gc3libs.log.debug("StagedTaskCollection '%s' has no stage%d,"
                              " ending sequence now.", self, (done + 1))
            self.execution.returncode = self.tasks[done].execution.returncode
            return Run.State.TERMINATED
        # get next stage (2); if we get an error here, something is wrong in
        # the code
        try:
            next_stage = next_stage_fn()
        except AttributeError as err:
            raise AssertionError(
                "Invalid `StagedTaskCollection` instance %r: %s"
                % (self, str(err)))
        # add next stage to the collection, or end graciously
        if isinstance(next_stage, Task):
            self.add(next_stage)
            return Run.State.RUNNING
        elif isinstance(next_stage, (int, int, tuple)):
            self.execution.returncode = next_stage
            return Run.State.TERMINATED
        else:
            raise AssertionError(
                "Invalid return value from method `stage%d()` of"
                " `StagedTaskCollection` object %r: must return `Task`"
                " instance or number" % (done + 1, self))


class ParallelTaskCollection(TaskCollection):

    """
    A `ParallelTaskCollection` runs all of its tasks concurrently.

    The collection state is set to `TERMINATED` once all tasks have
    reached the same terminal status.
    """

    def __init__(self, tasks=None, **extra_args):
        TaskCollection.__init__(self, tasks, **extra_args)

    def _state(self):
        """
        Return the state of the collection.

        For a `ParallelTaskCollection`, the state of dependent jobs is
        computed by looping across the states STOPPED, RUNNING,
        SUBMITTED, TERMINATING, UNKNOWN, TERMINATED, NEW in the order
        given: the first state for which there is at least one job in
        that state is returned as the overall collection state.  As an
        exception, if the collection is a mixture of NEW and
        TERMINATED jobs, then the global state is RUNNING (presuming
        we're in the middle of a computation).
        """
        stats = self.stats()
        # go through "live" states in order of "importance" and set
        # the full collection state to the first one that has tasks
        for state in [
                # special/error states of child task:
                Run.State.STOPPED,
                Run.State.UNKNOWN,
                # normal course of action:
                Run.State.RUNNING,
                Run.State.SUBMITTED,
        ]:
            if stats[state] > 0:
                return state
        # if we get here, then all jobs are all TERMINATING,
        # TERMINATED, or NEW; if there are any NEW ones, then we're in
        # the middle of a computation so `RUNNING` is the correct
        # state to report
        if stats[Run.State.NEW] > 0:
            return Run.State.RUNNING
        # `TERMINATING` wins over `TERMINATED`
        if stats[Run.State.TERMINATING] > 0:
            return Run.State.TERMINATING
        if stats[Run.State.TERMINATED] > 0:
            return Run.State.TERMINATED
        if self.tasks:
            gc3libs.log.error(
                "BUG! Non-empty task collection %r,"
                " yet `self.stats()` returned zero...", self)
        else:
            gc3libs.log.warning(
                "update_state() called on empty task collection %s"
                " -- this operation makes no sense,"
                " setting collection state to `TERMINATED`", self)
        return Run.State.TERMINATED

    def add(self, task):
        """
        Add a task to the collection.
        """
        self.tasks.append(task)
        if self._attached:
            task.attach(self._controller)
        else:
            task.detach()

    def attach(self, controller):
        """
        Use the given Controller interface for operations on the job
        associated with this task.
        """
        for task in self.tasks:
            if not task._attached:
                task.attach(controller)
        Task.attach(self, controller)

    def kill(self, **extra_args):
        """
        Terminate all tasks in the collection, and set collection
        state to `TERMINATED`.
        """
        for task in self.tasks:
            task.kill(**extra_args)
        self.execution.state = Run.State.TERMINATED
        self.execution.returncode = (Run.Signals.Cancelled, -1)
        self.changed = True

    def progress(self):
        """
        Try to advance all jobs in the collection to the next state in
        a normal lifecycle.
        """
        for task in self.tasks:
            task.progress()
        super(ParallelTaskCollection, self).progress()


    def redo(self, *args, **kwargs):
        """
        Reset collection and all included tasks to state ``NEW``.

        If not all included tasks should are in a terminal state or
        ``NEW``, an `AssertionError` exception will be thrown.
        See also `Task.redo`:meth: for a listing of allowed run states
        when ``redo()`` is called.
        """
        for task in self.tasks:
            task.redo(*args, **kwargs)
        super(ParallelTaskCollection, self).redo(*args, **kwargs)


    def submit(self, resubmit=False, targets=None, **extra_args):
        """
        Start all tasks in the collection.
        """
        one_submitted = False
        for task in self.tasks:
            try:
                task.submit(resubmit, targets, **extra_args)
                one_submitted = True
            except (gc3libs.exceptions.ResourceNotReady,
                    gc3libs.exceptions.MaximumCapacityReached):
                if one_submitted:
                    # stop submitting tasks
                    break
                else:
                    raise
        self.execution.state = self._state()


    def terminated(self):
        """
        Set exitcode based on termination status of sub-tasks.
        """
        self.execution.returncode = (0, 0)
        for task in self.tasks:
            if task.execution.returncode != 0:
                self.execution.exitcode = os.EX_SOFTWARE


    def update_state(self, **extra_args):
        """
        Update state of all tasks in the collection.
        """
        super(ParallelTaskCollection, self).update_state()
        self.execution.state = self._state()


class ChunkedParameterSweep(ParallelTaskCollection):

    def __init__(self, min_value, max_value, step, chunk_size, **extra_args):
        """
        Like `ParallelTaskCollection`, but generate a sequence of jobs
        with a parameter varying from `min_value` to `max_value` in
        steps of `step`.  Only `chunk_size` jobs are generated at a
        time, to distribute the burden of job creation along the whole run.
        """
        self.min_value = min_value
        self.max_value = max_value
        self.step = step
        self.chunk_size = chunk_size
        self._floor = min(min_value + (chunk_size * step), max_value)
        initial = list()
        for param in range(min_value, self._floor, step):
            initial.append(self.new_task(param))
        # start with the initial chunk of jobs
        ParallelTaskCollection.__init__(self, initial, **extra_args)

    def new_task(self, param, **extra_args):
        """
        Return the `Task` corresponding to the parameter value `param`.

        This method *must* be overridden in subclasses to generate tasks.
        """
        raise NotImplementedError(
            "Abstract method `ChunkedParameterSweep.new_task()` called -"
            " this should have been defined in a derived class.")

    # this is called at every cycle
    def update_state(self, **extra_args):
        """
        Like `ParallelTaskCollection.update_state()`,
        but also creates new tasks if less than
        `chunk_size` are running.
        """
        ParallelTaskCollection.update_state(self, **extra_args)
        # XXX: proposal, reset chuck_size from self._controller.max_in_flight
        # this is the way to pass new 'max-running' value to the class
        # this creates though, a tigh coupling with 'controller' and maybe
        # limits the flexibility of the class.
        # In this way we obsolete 'chunked_size' as part of the __init__ args
        # if self._controller:
        #     gc3libs.log.info("Updating %s chunk_size from %d to %d" %
        #                      (self.__class__,
        #                       self.chunk_size,
        #                       self._controller.max_in_flight))
        #     self.chunk_size =  self._controller.
        # XXX: shall we als could jobs in Run.State.STOPPED ?
        num_running = len([task for task in self.tasks if
                           task.execution.state in [Run.State.NEW,
                                                    Run.State.SUBMITTED,
                                                    Run.State.RUNNING]])
        # Run.State.UNKNOWN ]])
        # add more jobs if we're close to the end
        # XXX: why using 2*self.chunk_size as treshold ?
        # is the idea to submit more jobs once we reach at least 50%
        # of completion ?
        # if num_running < 2*self.chunk_size and self._floor < self.max_value:
        if 2 * num_running < self.chunk_size and self._floor < self.max_value:
            # generate more tasks
            top = min(
                self._floor + (self.chunk_size * self.step), self.max_value)
            for param in range(self._floor, top, self.step):
                self.add(self.new_task(param, **extra_args))
            self._floor = top
            self.execution.state = self._state()
            self.changed = True
        return self.execution.state


class RetryableTask(Task):

    """
    Wrap a `Task` instance and re-submit it until a specified
    termination condition is met.

    By default, the re-submission upon failure happens iff execution
    terminated with nonzero return code; the failed task is retried up
    to `self.max_retries` times (indefinitely if `self.max_retries` is 0).

    Override the `retry` method to implement a different retryal policy.

    *Note:* The resubmission code is implemented in the
    `terminated`:meth:, so be sure to call it if you override in
    derived classes.
    """

    def __init__(self, task, max_retries=0, **extra_args):
        """
        Wrap `task` and resubmit it until `self.retry()` returns `False`.

        :param Task task: A `Task` instance that should be retried.

        :param int max_retries: Maximum number of times `task` should be
            re-submitted; use 0 for 'no limit'.
        """
        self.max_retries = max_retries
        self.retried = 0
        self.task = task
        self.would_output = self.task.would_output
        Task.__init__(self, **extra_args)

    @property
    def changed(self):
        """
        Evaluates to `True` if this task or any of its subtasks has been
        modified and should be saved to persistent storage.
        """
        return self._changed or self.task.changed

    @changed.setter
    def changed(self, value):
        self._changed = value

    def __getattr__(self, name):
        """Proxy public attributes of the wrapped task."""
        if name.startswith('_'):
            raise AttributeError(
                "'%s' object has no attribute '%s'"
                % (self.__class__.__name__, name))
        return getattr(self.task, name)

    def retry(self):
        """
        Return `True` or `False`, depending on whether the failed task
        should be re-submitted or not.

        The default behavior is to retry a task iff its execution
        terminated with nonzero returncode and the maximum retry limit
        has not been reached.  If `self.max_retries` is 0, then the
        dependent task is retried indefinitely.

        Override this method in subclasses to implement a different
        policy.
        """
        if (self.task.execution.returncode != 0
            and ((self.max_retries > 0
                  and self.retried < self.max_retries)
                 or self.max_retries == 0)):
            return True
        else:
            return False

    def attach(self, controller):
        # here `Task.attach` is the invocation of the superclass'
        # `attach` method (which attaches *this* object to a controller),
        # while `self.task.attach` is the propagation of the `attach`
        # method to the wrapped task. (Same for `detach` below.)
        Task.attach(self, controller)
        self.task.attach(controller)

    def detach(self):
        # see comment in `attach` above
        Task.detach(self)
        self.task.detach()

    def fetch_output(self, *args, **extra_args):
        self.task.fetch_output(*args, **extra_args)

    def free(self, **extra_args):
        self.task.free(**extra_args)

    def kill(self, **extra_args):
        self.task.kill(**extra_args)
        if self.task.execution.state == Run.State.TERMINATED:
            self.execution.state = Run.State.TERMINATED
            self.changed = True

    def peek(self, *args, **extra_args):
        return self.task.peek(*args, **extra_args)

    def submit(self, resubmit=False, targets=None, **extra_args):
        # XXX: previous code ignored the `resubmit` argument; why?!
        self.task.submit(resubmit, targets, **extra_args)
        # immediately update state if submission of managed task was
        # successful; otherwise this task may remain in ``NEW`` state
        # which causes an unwanted resubmission if the managing
        # programs ends or is interrupted just after the submission...
        # XXX: this is a case for a generic publish/subscribe mechanism!
        if self.task.execution.state != Run.State.NEW:
            self.execution.state = self._recompute_state()

    def _recompute_state(self):
        """
        Determine and return the state based on the current state and
        the state of the wrapped task.
        """
        own_state = self.execution.state
        task_state = self.task.execution.state
        if own_state == task_state:
            return own_state
        elif own_state == Run.State.NEW:
            if task_state == Run.State.NEW:
                return Run.State.NEW
            elif task_state in [Run.State.SUBMITTED,
                                Run.State.RUNNING,
                                Run.State.STOPPED,
                                Run.State.UNKNOWN]:
                return task_state
            else:
                return Run.State.RUNNING
        elif own_state == Run.State.SUBMITTED:
            if task_state in [Run.State.NEW, Run.State.SUBMITTED]:
                return Run.State.SUBMITTED
            elif task_state in [Run.State.RUNNING,
                                Run.State.TERMINATING,
                                Run.State.TERMINATED]:
                return Run.State.RUNNING
            else:
                return task_state
        elif own_state == Run.State.RUNNING:
            if task_state in [Run.State.STOPPED, Run.State.UNKNOWN]:
                return task_state
            else:
                # if task is NEW, SUBMITTED, RUNNING, etc. -- keep our state
                return own_state
        elif own_state in [Run.State.TERMINATING, Run.State.TERMINATED]:
            assert task_state == Run.State.TERMINATED
            return Run.State.TERMINATED
        elif own_state in [Run.State.STOPPED, Run.State.UNKNOWN]:
            if task_state in [Run.State.NEW,
                              Run.State.SUBMITTED,
                              Run.State.RUNNING,
                              Run.State.TERMINATING,
                              Run.State.TERMINATED]:
                return Run.State.RUNNING
            else:
                return own_state
        else:
            # should not happen!
            raise AssertionError(
                "Unhandled own state '%s' in RetryableTask._recompute_state()",
                own_state)

    def update_state(self):
        """
        Update the state of the dependent task, then resubmit it if it's
        TERMINATED and `self.retry()` is `True`.
        """
        own_state_old = self.execution.state
        if self.task.execution.state not in [Run.State.NEW, Run.State.TERMINATED]:
            self.task.update_state()
        own_state_new = self._recompute_state()
        if (self.task.execution.state == Run.State.TERMINATED
                and own_state_old != Run.State.TERMINATED):
            self.execution.returncode = self.task.execution.returncode
            if self.retry():
                self.retried += 1
                self.task.submit(resubmit=True)
                own_state_new = Run.State.RUNNING
            else:
                own_state_new = Run.State.TERMINATED
            self.changed = True
        if own_state_new != own_state_old:
            self.execution.state = own_state_new
            self.changed = True


class DependentTaskCollection(SequentialTaskCollection):

    """
    Run a set of tasks, respecting inter-dependencies between them.

    Each task can list a number of tasks that need to be run before
    it; upon submission, a `DependentTaskCollection` creates a direct
    acyclic graph from that dependency information and ensures that no
    task is run before its dependencies have been successfully
    executed.

    The collection state is set to `TERMINATED` once all tasks have
    reached the same terminal status.
    """

    def __init__(self, tasks=None, **extra_args):
        super(DependentTaskCollection, self).__init__([], **extra_args)
        # record what tasks were given, but only move them to the
        # actual execution list when *this* TaskCollection is first
        # submitted
        self._deps = defaultdict(set)
        if tasks:
            for task in tasks:
                self.add(task)

    def add(self, task, after=None):
        """
        Add a task to the collection.

        The task will be run after any tasks referenced in the `after`
        sequence have terminated their run.  Alternatively, a task can
        list tasks it depends upon in its ``.after`` attribute; i.e.,
        the following two syntaxes are equivalent:

            | >>> coll.add(task1, after=[task2])
            |
            | >>> task1.after = [task2]
            | >>> coll.add(task1)

        **Note:** tasks can only be added to a
        `DependentTaskCollection` while it's in state ``NEW``.
        """
        assert (self.execution.state == Run.State.NEW), \
            "Can only add tasks to a DependentTaskCollection while it's in state `NEW`"
        # collect all task dependencies
        task_dependencies = self._deps[task]
        if after is not None:
            task_dependencies.update(after)
        try:
            task_after = task.after
            if task_after:
                task_dependencies.update(task_after)
        except AttributeError:
            pass

    def submit(self, resubmit=False, targets=None, **extra_args):
        if self.execution.state == Run.State.NEW:
            extras = {'would_output': self.would_output}
            if 'output_dir' in self:
                extras['output_dir'] = self.output_dir
            # create DAG from dependency information
            sorted_and_grouped_tasks = toposort(self._deps)
            for batch in sorted_and_grouped_tasks:
                step = ParallelTaskCollection(batch, **extras)
                super(DependentTaskCollection, self).add(step)
        super(DependentTaskCollection, self).submit(resubmit, targets, **extra_args)


# main: run tests

if "__main__" == __name__:
    import doctest
    doctest.testmod(name="dag",
                    optionflags=doctest.NORMALIZE_WHITESPACE)



#! /usr/bin/env python
#
#   gtraclong.py -- Front-end script for running the Tracula
#   over a list of subject files.
#
#   Copyright (C) 2016, 2017  University of Zurich. All rights reserved.
#
#   This program is free software: you can redistribute it and/or
#   modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""

It uses the generic `gc3libs.cmdline.SessionBasedScript` framework.

See the output of ``gtraclong.py --help`` for program usage
instructions.
"""

# summary of user-visible changes
__changelog__ = """
  2016-07-05:
  * Initial version
"""
__author__ = 'Sergio Maffioletti <sergio.maffioletti@uzh.ch>'
__docformat__ = 'reStructuredText'


# run script, but allow GC3Pie persistence module to access classes defined here;
# for details, see: https://github.com/uzh/gc3pie/issues/95
if __name__ == "__main__":
    import gtraclong
    gtraclong.GtraclongScript().run()

import os
import sys
import time
import tempfile
import re

import shutil
# import csv

from pkg_resources import Requirement, resource_filename

import gc3libs
import gc3libs.exceptions
from gc3libs import Application, Run, Task
from gc3libs.cmdline import SessionBasedScript, executable_file
import gc3libs.utils
from gc3libs.quantity import Memory, kB, MB, MiB, GB, Duration, hours, minutes, seconds
from gc3libs.workflow import RetryableTask

DEFAULT_CORES = 1
DEFAULT_MEMORY = Memory(3000,MB)

DEFAULT_REMOTE_INPUT_FOLDER="./"
DEFAULT_REMOTE_OUTPUT_FOLDER="./output"
DMRIC_PATTERN = "dmrirc"
DEFAULT_TRAC_COMMAND="trac-all -prep -c {dmrirc} -debug"

def _correlate(subjects, freesurfers):
    """
    correlate every subject in 'subjects' with subfolders in 'freesurfers'
    """    
    for subject in subjects:
        freesurfer_list = list()
        for freesurfer in os.listdir(freesurfers):
            if freesurfer.split('.')[0] == subject:
                freesurfer_list.append(os.path.join(freesurfers,freesurfer))
        yield (subject,freesurfer_list)

## custom application class
class GtraclongApplication(Application):
    """
    """
    application_name = 'gtraclong'
    
    def __init__(self, subject, subject_folder, **extra_args):

        self.output_dir = extra_args['output_dir']

        inputs = dict()
        outputs = dict()

        inputs[subject_folder] = DEFAULT_REMOTE_INPUT_FOLDER

        wrapper = resource_filename(Requirement.parse("gc3pie"),
                                    "gc3libs/etc/gtraclong_wrapper.py")
        inputs[wrapper] = os.path.basename(wrapper)

        arguments = "./%s" % (inputs[wrapper])

        if extra_args['requested_memory'] < DEFAULT_MEMORY:
            gc3libs.log.warning("GtraclongApplication for subject %s running with memory allocation " \
                                "'%d GB' lower than suggested one: '%d GB'," % (subject,
                                                                                extra_args['requested_memory'].amount(unit=GB),
                                                                                DEFAULT_MEMORY.amount(unit=GB)))
        Application.__init__(
            self,
            arguments = arguments,
            inputs = inputs,
            outputs = [DEFAULT_REMOTE_OUTPUT_FOLDER],
            stdout = 'gtraclong.log',
            join=True,
            **extra_args)        

class GtraclongScript(SessionBasedScript):
    """
    
    The ``gtraclong`` command keeps a record of jobs (submitted, executed
    and pending) in a session file (set name with the ``-s`` option); at
    each invocation of the command, the status of all recorded jobs is
    updated, output from finished jobs is collected, and a summary table
    of all known jobs is printed.
    
    Options can specify a maximum number of jobs that should be in
    'SUBMITTED' or 'RUNNING' state; ``gtraclong`` will delay submission of
    newly-created jobs so that this limit is never exceeded.
    """

    def __init__(self):
        SessionBasedScript.__init__(
            self,
            version = __version__, # module version == script version
            application = GtraclongApplication, 
            # only display stats for the top-level policy objects
            # (which correspond to the processed files) omit counting
            # actual applications because their number varies over
            # time as checkpointing and re-submission takes place.
            stats_only_for = GtraclongApplication,
            )
 
    # def setup_args(self):

    #     self.add_param('input_data', type=str,
    #                    help="Root location of input data. "
    #                    "Note: expected folder structure: "
    #                    " 1 subfodler for each subject. "
    #                    " In each subject folder, " 
    #                    " 1 subfolder for each TimePoint. "
    #                    " Each TimePoint folder should contain 2 input "
    #                    "NFTI files.")

    def setup_options(self):
        self.add_param("-F", "--fs", metavar="[PATH]", 
                       dest="freesurfer", default=None,
                       help="Location of Freesurfer data folders.")
    
    def parse_args(self):
        self.params.subjects.append = list()
        try:
            for input_folder in self.params.args:
                assert os.path.isdir(input_folder), \
                    "Input subject forler %s not found" % input_folder
                self.params.subjects.append(input_folder)
        except ValueError as vx:
            gc3libs.log.warning(vx)
    
    def new_tasks(self, extra):
        """
        For each input folder, create an instance of GtraclongApplication
        """
        tasks = []
        for (subject_name,freesurfers) in _correlate(self.params.subjects, self.params.freesurfer)
        
            # extract root folder name to be used as jobname
            extra_args = extra.copy()
            extra_args['jobname'] = subject_name

            extra_args['output_dir'] = self.params.output
            extra_args['output_dir'] = extra_args['output_dir'].replace('NAME', 
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('SESSION', 
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('DATE', 
                                                                        'run_%s' % subject_name)
            extra_args['output_dir'] = extra_args['output_dir'].replace('TIME', 
                                                                        'run_%s' % subject_name)
            
            tasks.append(GtraclongApplication(
                subject_name,
                os.path.join(self.params.subjects,subject_name),
                freesurfers,
                **extra_args))
            
        return tasks

    def get_input_subject_folder(self, input_folder):
        """
        Check and validate input subfolders
        """

        for r,d,f in os.walk(input_folder):
            for infile in f:
                if infile.startswith(DMRIC_PATTERN):
                    yield (os.path.abspath(r),os.path.basename(r),infile)

